#!/usr/bin/env python

#
# $File: __init__.py $
# $LastChangedDate$
# $Rev$
#
# This file is part of simuPOP, a forward-time population genetics
# simulation environment. Please visit http://simupop.sourceforge.net
# for details.
#
# Copyright (C) 2004 - 2010 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
simuPOP core module.
"""

__all__ = [
    '__version__',
    # Constants
    #
    'MALE',
    'FEMALE',
    #
    'CUSTOMIZED',
    'AUTOSOME',
    'CHROMOSOME_X',
    'CHROMOSOME_Y',
    'MITOCHONDRIAL',
    #
    'CONSTANT',
    'BINOMIAL_DISTRIBUTION',
    'EXPONENTIAL_DISTRIBUTION',
    'GEOMETRIC_DISTRIBUTION',
    'POISSON_DISTRIBUTION',
    'UNIFORM_DISTRIBUTION',
    'NORMAL_DISTRIBUTION',
    'GAMMA_DISTRIBUTION',
    #
    'NO_SEX',
    'RANDOM_SEX',
    'PROB_OF_MALES',
    'NUM_OF_MALES',
    'NUM_OF_FEMALES',
    'SEQUENCE_OF_SEX',
    'GLOBAL_SEQUENCE_OF_SEX',
    #
    'NO_CONVERSION',
    'NUM_MARKERS',
    'TRACT_LENGTH',
    #
    'SPOUSE',
    'OUTBRED_SPOUSE',
    'FULLSIBLING',
    'SIBLING',
    'OFFSPRING',
    'COMMON_OFFSPRING',
    #
    'ANY_SEX',
    'MALE_ONLY',
    'FEMALE_ONLY',
    'SAME_SEX',
    'OPPOSITE_SEX',
    'PAIR_ONLY',
    #
    'MULTIPLICATIVE',
    'ADDITIVE',
    'HETEROGENEITY',
    #
    'BY_IND_INFO',
    'BY_PROBABILITY',
    'BY_PROPORTION',
    'BY_COUNTS',
    #
    'PATERNAL',
    'MATERNAL',
    #
    'MEAN',
    'MAXIMUM',
    'MINIMUM',
    'SUMMATION',
    'MULTIPLICATION',
    # 
    'PER_LOCI',
    'PER_CHROMOSOME',
    'PER_PLOIDY',
    'PER_INDIVIDUAL',
    'FROM_INFO',
    'FROM_INFO_SIGNED',
    #
    'HAPLODIPLOID',
    'ALL_AVAIL',
    'UNSPECIFIED',
    # 
    # type
    'defdict',
    #
    # Major simuPOP classes
    'Population',
    # This is just to make help(Individual) available to users.
    'Individual',
    'Simulator',
    'Pedigree',
    # splitters
    'SexSplitter',
    'AffectionSplitter',
    'CombinedSplitter',
    'ProductSplitter',
    'ProportionSplitter',
    'InfoSplitter',
    'GenotypeSplitter',
    'RangeSplitter',
    # mating schemes
    'MatingScheme',
    'HomoMating',
    'HeteroMating',
    'ConditionalMating',
    'PedigreeMating',
    'OffspringGenerator',
    'ControlledOffspringGenerator',
    'RandomParentChooser',
    'PyParentsChooser',
    'CombinedParentsChooser',
    'SequentialParentChooser',
    'RandomParentsChooser',
    'SequentialParentsChooser',
    'ParentChooser',
    'PolyParentsChooser',
    #
    'RandomMating',
    'RandomSelection',
    'MonogamousMating',
    'SelfMating',
    'HermaphroditicMating',
    'CloneMating',
    'HaplodiploidMating',
    #'consanguineousMating',
    'ControlledRandomMating',
    'PolygamousMating',
    #
    #
    # Operators
    #
    'InitSex',
    'InitInfo',
    'InitGenotype',
    'InitLineage',
    #
    'PyOutput',
    'PyEval',
    'PyExec',
    'InfoEval',
    'InfoExec',
    #
    'Migrator',
    'BackwardMigrator',
    'MergeSubPops',
    'SplitSubPops',
    'ResizeSubPops',
    #
    'GenoTransmitter',
    'HaplodiploidGenoTransmitter',
    'MendelianGenoTransmitter',
    'MitochondrialGenoTransmitter',
    'SelfingGenoTransmitter',
    'CloneGenoTransmitter',
    'Recombinator',
    # 
    'PointMutator',
    'MatrixMutator',
    'MixedMutator',
    'PyMutator',
    'StepwiseMutator',
    'SNPMutator',
    'AcgtMutator',
    'AminoAcidMutator',
    'ContextMutator',
    'KAlleleMutator',
    'RevertFixedSites',
    'FiniteSitesMutator',
    #
    'MapSelector',
    'MaSelector',
    'MlSelector',
    'PySelector',
    'PyMlSelector',
    #
    'MaPenetrance',
    'MapPenetrance',
    'MlPenetrance',
    'PyPenetrance',
    'PyMlPenetrance',
    #
    'PyQuanTrait',
    #
    'Stat',
    #
    'IdTagger',
    'InheritTagger',
    'ParentsTagger',
    'OffspringTagger',
    'PedigreeTagger',
    'PyTagger',
    'SummaryTagger',
    #
    'TerminateIf',
    'RevertIf',
    'DiscardIf',
    'discardIf',
    #
    'PyOperator',
    #
    'NoneOp',
    'Dumper',
    'SavePopulation',
    'IfElse',
    'Pause',
    'TicToc',
    #
    # Function form of operators
    'dump',
    #
    'initSex',
    'initInfo',
    'initGenotype',
    'initLineage',
    #
    'pyEval',
    'pyExec',
    'infoEval',
    'infoExec',
    #
    'tagID',
    #
    'migrate',
    'backwardMigrate',
    'resizeSubPops',
    'splitSubPops',
    'mergeSubPops',
    #
    'matrixMutate',
    'contextMutate',
    'kAlleleMutate',
    'mixedMutate',
    'pointMutate',
    'pyMutate',
    'stepwiseMutate',
    'snpMutate',
    'acgtMutate',
    'revertFixedSites',
    #
    'stat',
    #
    # Global functions
    'WithArgs',
    'WithMode',
    # RNG Related
    'RNG',
    'getRNG',
    'setRNG',
    #
    'closeOutput',
    'describeEvolProcess',
    'loadPopulation',
    'loadPedigree',
    'moduleInfo',
    'turnOffDebug',
    'turnOnDebug',
    'setOptions',
    #
    'maPenetrance',
    'mapPenetrance',
    'mlPenetrance',
    'pyPenetrance',
    #
    'pyQuanTrait',
    #
    # For testing only
    'applyDuringMatingOperator',
    'Bernullitrials',
    'Bernullitrials_T',
    'WeightedSampler',
    # 
    # modules are not loaded by default because they require rpy or matplotlib
    #
    #'utils',
    #'sampling',
]

# get options
from simuOpt import simuOptions
import os, sys, re

if simuOptions['Optimized']:
    if simuOptions['AlleleType'] == 'short':
        from simuPOP.simuPOP_op import *
    elif simuOptions['AlleleType'] == 'long':
        from simuPOP.simuPOP_laop import *
    elif simuOptions['AlleleType'] == 'binary':
        from simuPOP.simuPOP_baop import *
    elif simuOptions['AlleleType'] == 'mutant':
        from simuPOP.simuPOP_muop import *
    elif simuOptions['AlleleType'] == 'lineage':
        from simuPOP.simuPOP_linop import *
    else:
        from simuPOP.simuPOP_op import *
else:
    if simuOptions['AlleleType'] == 'short':
        from simuPOP.simuPOP_std import *
    elif simuOptions['AlleleType'] == 'long':
        from simuPOP.simuPOP_la import *
    elif simuOptions['AlleleType'] == 'binary':
        from simuPOP.simuPOP_ba import *
    elif simuOptions['AlleleType'] == 'mutant':
        from simuPOP.simuPOP_mu import *
    elif simuOptions['AlleleType'] == 'lineage':
        from simuPOP.simuPOP_lin import *
    else:
        from simuPOP.simuPOP_std import *

__version__ = moduleInfo()['version']

if simuOptions['Version'] is not None:
    expMajor, expMinor, expRelease = [int(x) for x in re.match(r'^(\d+)\.(\d+)\.(\d+)', simuOptions['Version']).groups()]
    myMajor, myMinor, myRelease = [int(x) for x in re.match(r'^(\d+)\.(\d+)\.(\d+)', __version__).groups()]
    if (expMajor > myMajor) or (expMajor == myMajor and expMinor > myMinor) or \
        (expMajor == myMajor and expMinor == myMinor and expRelease > myRelease):
        raise ImportError('simuPOP version %s is installed but version >= %s is required. ' % \
            (__version__, simuOptions['Version']) + 
            'Please upgrade your simuPOP installation.')

if simuOptions['Revision'] is not None:
    rev = moduleInfo()['revision']
    if rev < simuOptions['Revision']:
        raise ImportError('simuPOP version %s (revision %d) is installed ' % (__version__, rev) +
            'but simuPOP revision >= %d is required. ' % simuOptions['Revision'] +
            'Please upgrade your simuPOP installation.')
    if 'DBG_COMPATIBILITY' in simuOptions['Debug']:
        print('WARNING: parameter revision in simuOpt.setOptions is obsolete and might be removed from a future version of simuPOP.')

# set number of threads in openMP
if simuOptions['NumThreads'] is not None:
    setOptions(numThreads=simuOptions['NumThreads'])

if not simuOptions['Quiet']:
    info = moduleInfo()
    print("simuPOP Version %s : Copyright (c) 2004-2016 Bo Peng" % (__version__))
    # compile date, compiler etc are macros that are replaced during compile time.
    print("Revision %d (%s) for Python %s (%dbit, %d%s)" % \
            (info['revision'], info['date'], info['python'], info['wordsize'], info['threads'],
                    'thread' if info['threads'] <= 1 else 'threads'))
    print("Random Number Generator is set to %s with random seed 0x%08x." % (getRNG().name(), getRNG().seed()))
    # MaxAllele + 1 since 0 is one of the allelic states
    print("This is the %s %s allele version with %d maximum allelic states." % \
            ('optimized' if info['optimized'] else 'standard', info['alleleType'], info['maxAllele']+1))
    print("For more information, please visit http://simupop.sourceforge.net,")
    print("or email simupop-list@lists.sourceforge.net (subscription required).")
    # Turn on general debug information when not in 'quiet' mode
    # This will print out error messages when needed.
    turnOnDebug('DBG_GENERAL')
    # only display banner once
    simuOptions['Quiet'] = True

if simuOptions['Debug'] != []:
    for g in simuOptions['Debug']:
        if g not in ['', None]:
            print("Turn on debug '%s'" % g)
            turnOnDebug(g)


ALL_AVAIL = True
UNSPECIFIED = False

def evolve_pop(self, initOps=[], preOps=[], matingScheme=MatingScheme(), postOps=[],
    finalOps=[], gen=-1, dryrun=False):
    '''Evolve the current population *gen* generations using mating scheme
    *matingScheme* and operators *initOps* (applied before evolution), *preOps*
    (applied to the parental population at the beginning of each life cycle),
    *postOps* (applied to the offspring population at the end of each life
    cycle) and *finalOps* (applied at the end of evolution). More specifically,
    this function creates a *Simulator* using the current population, call its
    *evolve* function using passed parameters and then replace the current
    population with the evolved population. Please refer to function 
    ``Simulator.evolve`` for more details about each parameter.'''
    if dryrun:
        print(describeEvolProcess(initOps, preOps, matingScheme, postOps, finalOps, gen, 1))
        return (0,)
    if isinstance(self, Pedigree):
        raise ValueError("Evolving a pedigree object directly is not allowed.")
    # create a simulator with self
    simu = Simulator(self)
    # evolve
    gen = simu.evolve(initOps, preOps, matingScheme, postOps, finalOps, gen)
    # get the evolved population
    self.swap(simu.population(0))
    return gen[0]

Population.evolve = evolve_pop

def all_individuals(self, subPops=ALL_AVAIL, ancGens=ALL_AVAIL):
    '''Return an iterator that iterat through all (virtual) subpopulations in
    all ancestral generations. A list of (virtual) subpopulations (*subPops*)
    and a list of ancestral generations (*ancGens*, can be a single number)
    could be specified to iterate through only selected subpopulation and
    generations. Value ``ALL_AVAIL`` is acceptable in the specification of
    ``sp`` and/or ``vsp`` in specifying a virtual subpopulation ``(sp, vsp)``
    for the iteration through all or specific virtual subpopulation in all or
    specific subpopulations.
    '''
    if ancGens is ALL_AVAIL:
        gens = range(self.ancestralGens() + 1)
    elif hasattr(ancGens, '__iter__'):
        gens = ancGens
    else:
        gens = [ancGens]
    #
    curGen = self.curAncestralGen()
    for gen in gens:
        self.useAncestralGen(gen)
        if subPops is ALL_AVAIL:
            for ind in self.individuals():
                yield ind
        else:
            for subPop in subPops:
                if hasattr(subPop, '__iter__'):
                    if len(subPop) != 2:
                        raise ValueError('Invalid subpopulation ID %s' % subPop)
                    if subPop[0] is ALL_AVAIL:
                        # (ALL_AVAIL, ALL_AVAIL)
                        if subPop[1] is ALL_AVAIL and self.numVirtualSubPop() > 0:
                            for sp in range(self.numSubPop()):
                                for vsp in range(self.numVirtualSubPop()):
                                    for ind in self.individuals([sp, vsp]):
                                        yield ind
                        # (ALL_AVAIL, vsp)
                        else:
                            for sp in range(self.numSubPop()):
                                for ind in self.individuals([sp, subPop[1]]):
                                    yield ind
                    else:
                        # (sp, ALL_AVAIL)
                        if subPop[1] is ALL_AVAIL and self.numVirtualSubPop() > 0:
                                for vsp in range(self.numVirtualSubPop()):
                                    for ind in self.individuals([subPop[0], vsp]):
                                        yield ind
                        # (sp, vsp)
                        else:
                            for ind in self.individuals(subPop):
                                yield ind
                else:
                    for ind in self.individuals(subPop):
                        yield ind
    self.useAncestralGen(curGen)

Population.allIndividuals = all_individuals

def as_pedigree(self, idField='ind_id', fatherField='father_id', motherField='mother_id'):
    '''Convert the existing population object to a pedigree. After this function
    pedigree function should magically be usable for this function.
    '''
    if isinstance(self, Pedigree):
        return
    ped = Pedigree(self, loci=ALL_AVAIL, infoFields=ALL_AVAIL, ancGens=ALL_AVAIL,
        idField=idField, fatherField=fatherField, motherField=motherField,
        stealPop=True)
    # swap ped and this object. (I do not know if this is the right thing to do)
    self.__class__, ped.__class__ = ped.__class__, self.__class__
    self.this, ped.this = ped.this, self.this

Population.asPedigree = as_pedigree

def as_population(self):
    '''Convert the existing pedigree object to a population. This function will
    behave like a regular population after this function call.'''
    if isinstance(self, Population):
        return
    pop = population(0)
    # the pedigree data has been swapped to pop
    pop.swap(self)
    # swap ped and this object. (I do not know if this is the right thing to do)
    self.__class__, pop.__class__ = pop.__class__, self.__class__
    self.this, pop.this = pop.this, self.this

Pedigree.asPopulation = as_population

# Other definitions that does not really belong to simuUtil.py
class _dw(object):
    def __init__(self, var):
        try:
            self.__dict__ = var
        except TypeError:
            raise TypeError("The returned value is not a dictionary.\nNote: simu.vars() is a list so simu.dvars() is not allowed. \n    Use simu.dvars(rep) for population namespace.")
    def clear(self):
        self.__dict__.clear()
    def __repr__(self):
        return str(self.__dict__)

def dvars(self, *args, **kwargs):
    return _dw(self.vars(*args, **kwargs))

Population.dvars = dvars
Simulator.dvars = dvars

# expose the clone() method to Python copy module.
def _deepcopy(self, memo):
    return self.clone()

Population.__deepcopy__ = _deepcopy
Simulator.__deepcopy__ = _deepcopy
BaseOperator.__deepcopy__ = _deepcopy


Population.__deepcopy__ = _deepcopy
Simulator.__deepcopy__ = _deepcopy
BaseOperator.__deepcopy__ = _deepcopy

def ind_setInfo2(self, field, value):
    self.setInfo(value, field)

def ind_setInfo3(self, field, value):
    if field == 'this':
        self.__dict__['this'] = value
    else:
        self.setInfo(value, field)

def ind_getInfo3(self, field):
    if field == 'this':
        return self.__dict__['this']
    else:
        return self.info(field)

if sys.version_info[0] == 3:
    Individual.__setattr__ = ind_setInfo3
    Individual.__getattr__ = ind_getInfo3
else:
    Individual.__setattr__ = ind_setInfo2
    Individual.__getattr__ = Individual.info


def obj_equal(self, obj):
    return self.__cmp__(obj) == 0

if sys.version_info[0] == 3:
    Population.__eq__ = obj_equal
    Pedigree.__eq__ = obj_equal
    Simulator.__eq__ = obj_equal

##
## This is the result of an attemp to move describeEvolProcess from the C++
## level to the Python level so that it can correctly call the 'describe'
## function of a derived operator or mating scheme if they define a 'describe'
## function.
##
## This change has been temporarily reverted because many simuPOP components
## such as HomoMating, OffspringGenerator accepts such objects but cannot
## call the Python level 'describe' function correctly. Because this function
## can only describe customized pre or post mating operators and homogeneous
## mating schemes, it does not appear to be a good idea to expose (although
## still hidden) functions such as BaseOperator.applicability() to the Python
## interface.
## 

## def describeEvolProcess(initOps = [], preOps = [], matingScheme = None,
##     postOps = [], finalOps = [], gen = -1, numRep = 1):
##     '''This function takes the same parameters as ``Simulator.evolve`` and
##     output a description of how an evolutionary process will be executed. It is
##     recommended that you call this function if you have any doubt on how your
##     simulation will proceed.'''
##     allDesc = [''] * numRep
## 
##     # handle single inputs
##     if not hasattr(initOps, '__iter__'):
##         initOps = [initOps]
##     if not hasattr(preOps, '__iter__'):
##         preOps = [preOps]
##     if not hasattr(postOps, '__iter__'):
##         postOps = [postOps]
##     if not hasattr(finalOps, '__iter__'):
##         finalOps = [finalOps]
## 
##     for curRep in range(numRep):
##         desc = ''
##         if not initOps:
##             desc += 'No operator is used to initialize Population (initOps).\n'
##         else:
##             desc += 'Apply pre-evolution operators to the initial population (initOps).\n<ul>\n'
##             for op in initOps:
##                 desc += '<li>' + op.describe(False) + ' ' + op.applicability(True, False) + '\n'
##             desc += '</ul>\n'
##         if gen < 0:
##             desc += '\nEvolve a population indefinitely until an operator determines it.\n'
##         else:
##             desc += '\nEvolve a population for %s generations\n' % gen
##         desc += '<ul>\n'
##         if not preOps:
##             desc += '<li>No operator is applied to the parental generation (preOps).\n'
##         else:
##             desc += '<li>Apply pre-mating operators to the parental generation (preOps)\n<ul>\n'
##             for op in preOps:
##                 if op.isActive(curRep, 0):
##                     desc += '<li>' + op.describe(False) + ' ' + op.applicability() + '\n'
##             desc += '</ul>\n'
##         #
##         desc += '\n<li>Populate an offspring populaton from the parental population using mating scheme ' \
##             + matingScheme.describe(False) + '\n'
##         #
##         if not postOps:
##             desc += '\n<li>No operator is applied to the offspring population (postOps).\n'
##         else:
##             desc += '\n<li>Apply post-mating operators to the offspring population (postOps).\n<ul>\n'
##             for op in postOps:
##                 if op.isActive(curRep, 0):
##                     desc += '<li>' + op.describe(False) + ' ' + op.applicability() + '\n'
##             desc += '</ul>\n'
##         desc += '</ul>\n\n'
##         #
##         if not finalOps:
##             desc += 'No operator is applied to the final population (finalOps).\n'
##         else:
##             desc += 'Apply post-evolution operators (finalOps)\n<ul>\n'
##             for op in finalOps:
##                 desc += '<li>' + op.describe(False) + ' ' + op.applicability(True, False) + '\n'
##             desc += '</ul>\n'
##         #
##         allDesc[curRep] = desc
##     #
##     reps = []
##     for curRep in range(numRep):
##         if not reps:
##             reps.append(str(curRep))
##         else:
##             if allDesc[curRep] == allDesc[curRep - 1]:
##                 reps.append(str(curRep))
##             else:
##                 desc += 'Replicate' + ' '.join(reps)
##                 desc += ':\n' + allDesc[curRep - 1] + '\n'
##                 reps = [str(curRep)]
##     #
##     desc += 'Replicate' + ' '.join(reps) + ':\n' + allDesc[-1]
##     return formatDescription(desc)

class WithArgs:
    '''This class wraps around a user-provided function and provides an
    attribute ``args`` so that simuPOP knows which parameters to send to the
    function. This is only needed if the function can not be defined with
    allowed parameters.
    '''
    def __init__(self, func, args):
        '''Return a callable object that wraps around function ``func``.
        Parameter ``args`` should be a list of parameter names.
        '''
        self.__args__ = args
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

class WithMode:
    '''This class wraps around a user-provided output string, function
    or file handle (acceptable by parameter ``output`` of operators) so
    that simuPOP knows which mode the output should be written to. For
    example, if the output of the operator is a binary compressed stream,
    ``WithMode(output, 'b')`` could be used to tell the operators to 
    output bytes instead of string. This is most needed for Python 3 
    because files in Python 2 accepts string even if they are opened in
    binary mode.'''
    def __init__(self, output, mode=''):
        '''Return an object that wraps around ``output`` and tells simuPOP
        to output string in ``mode``. This class currently only support
        ``mode=''`` for text mode and ``mode='b'`` for binary output.'''
        self._with_output = output
        self._with_mode = mode


# mating schemes

class SequentialParentsChooser(CombinedParentsChooser):
    '''This parent chooser chooses two parents (a father and a mother)
    sequentially from their respective sex groups. Selection is not considered.
    If all fathers (or mothers) are exhausted, this parent chooser will choose
    fathers (or mothers) from the beginning of the (virtual) subpopulation
    again.'''
    def __init__(self):
        '''Create a parent chooser that chooses two parents sequentially from a
        parental (virtual) subpopulation.'''
        CombinedParentsChooser.__init__(self,
            SequentialParentChooser(MALE_ONLY),
            SequentialParentChooser(FEMALE_ONLY))


class CloneMating(HomoMating):
    '''A homogeneous mating scheme that uses a sequential parent chooser and
    a clone offspring generator.'''
    def __init__(self, numOffspring = 1, sexMode = None, ops = CloneGenoTransmitter(),
        subPopSize = [], subPops = ALL_AVAIL, weight = 0, selectionField = None):
        '''Create a clonal mating scheme that clones parents to offspring using
        a ``CloneGenoTransmitter``. Please refer to class ``OffspringGenerator``
        for parameters *ops* and *numOffspring*, and to class ``HomoMating`` for
        parameters  *subPopSize*, *subPops* and *weight*. Parameters *sexMode* and
        *selectionField* are ignored because this mating scheme does not support
        natural selection, and ``CloneGenoTransmitter`` copies sex from parents
        to offspring. Note that ``CloneGenoTransmitter`` by default also copies
        all parental information fields to offspring.
        '''
        HomoMating.__init__(self,
            chooser = SequentialParentChooser(),
            generator = OffspringGenerator(ops, numOffspring, RANDOM_SEX),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)


class RandomSelection(HomoMating):
    '''A homogeneous mating scheme that uses a random single-parent parent
    chooser with replacement, and a clone offspring generator. This mating
    scheme is usually used to simulate the basic haploid Wright-Fisher model
    but it can also be applied to diploid populations.'''
    def __init__(self, numOffspring = 1, sexMode = None, ops = CloneGenoTransmitter(),
        subPopSize = [], subPops = ALL_AVAIL, weight = 0, selectionField = 'fitness'):
        '''Create a mating scheme that select a parent randomly and copy him or
        her to the offspring population. Please refer to class 
        ``RandomParentChooser`` for parameter *selectionField*, to class
        ``OffspringGenerator`` for parameters *ops* and *numOffspring*, and to
        class ``HomoMating`` for parameters *subPopSize*, *subPops* and *weight*.
        Parameter *sexMode* is ignored because ``cloneOffspringGenerator`` copies
        sex from parents to offspring.
        '''
        HomoMating.__init__(self,
            chooser = RandomParentChooser(True, selectionField),
            generator = OffspringGenerator(ops, numOffspring, RANDOM_SEX),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)


class RandomMating(HomoMating):
    '''A homogeneous mating scheme that uses a random parents chooser with
    replacement and a Mendelian offspring generator. This mating scheme is
    widely used to simulate diploid sexual Wright-Fisher random mating.'''
    def __init__(self, numOffspring = 1, sexMode = RANDOM_SEX, ops = MendelianGenoTransmitter(), 
        subPopSize = [], subPops = ALL_AVAIL, weight = 0, selectionField = 'fitness'):
        '''Creates a random mating ssheme that selects two parents randomly and
        transmit genotypes according to Mendelian laws. Please refer to class
        ``RandomParentsChooser`` for parameter *selectionField*, to class
        ``OffspringGenerator`` for parameters *ops*, *sexMode* and
        *numOffspring*, and to class ``HomoMating`` for parameters
        *subPopSize*, *subPops* and *weight*.
        '''
        HomoMating.__init__(self,
            chooser = RandomParentsChooser(True, selectionField),
            generator = OffspringGenerator(ops, numOffspring, sexMode),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)


class MonogamousMating(HomoMating):
    '''A homogeneous mating scheme that uses a random parents chooser without
    replacement and a Mendelian offspring generator. It differs from the basic
    random mating scheme in that each parent can mate only once so there is no
    half-sibling in the population.'''
    def __init__(self, numOffspring = 1, sexMode = RANDOM_SEX, ops = MendelianGenoTransmitter(),
        subPopSize = [], subPops = ALL_AVAIL, weight = 0, selectionField = None):
        '''Creates a monogamous mating scheme that selects each parent only
        once. Please refer to class ``OffspringGenerator`` for parameters
        *ops*, *sexMode* and *numOffspring*, and to class ``HomoMating`` for
        parameters *subPopSize*, *subPops* and *weight*. Parameter
        *selectionField* is ignored because this mating scheme does not
        support natural selection.
        '''
        HomoMating.__init__(self,
            chooser = RandomParentsChooser(replacement=False),
            generator = OffspringGenerator(ops, numOffspring, sexMode),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)


class PolygamousMating(HomoMating):
    '''A homogeneous mating scheme that uses a multi-spouse parents chooser
    and a Mendelian offspring generator. It differs from the basic random
    mating scheme in that each parent of sex *polySex* will have *polyNum*
    spouses.'''
    def __init__(self, polySex=MALE, polyNum=1, numOffspring = 1,
        sexMode = RANDOM_SEX, ops = MendelianGenoTransmitter(), subPopSize = [],
        subPops = ALL_AVAIL, weight = 0, selectionField = 'fitness'):
        '''Creates a polygamous mating scheme that each parent mates with
        multiple spouses. Please refer to class ``PolyParentsChooser`` for
        parameters *polySex*, *polyNum* and *selectionField*, to class
        ``OffspringGenerator`` for parameters *ops*,  *sexMode* and
        *numOffspring*, and to class ``HomoMating`` for parameters
        *subPopSize*, *subPops* and *weight*. '''
        HomoMating.__init__(self,
            chooser = PolyParentsChooser(polySex, polyNum),
            generator = OffspringGenerator(ops, numOffspring, sexMode),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)


class HaplodiploidMating(HomoMating):
    '''A homogeneous mating scheme that uses a random parents chooser with
    replacement and a haplodiploid offspring generator. It should be used
    in a haplodiploid population where male individuals only have one set
    of homologous chromosomes.'''
    def __init__(self, numOffspring = 1., sexMode = RANDOM_SEX,
        ops = HaplodiploidGenoTransmitter(), subPopSize = [], subPops = ALL_AVAIL,
        weight = 0, selectionField = 'fitness'):
        '''Creates a mating scheme in haplodiploid populations. Please refer
        to class ``RandomParentsChooser`` for parameter *selectionField*, to
        class ``OffspringGenerator`` for parameters *ops*, *sexMode* and
        *numOffspring*, and to class ``HomoMating`` for parameters
        *subPopSize*, *subPops* and *weight*.
        '''
        HomoMating.__init__(self,
            chooser = RandomParentsChooser(True, selectionField),
            generator = OffspringGenerator(ops, numOffspring, sexMode),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)


class SelfMating(HomoMating):
    '''A homogeneous mating scheme that uses a random single-parent parent
    chooser with or without replacement (parameter *replacement*) and a
    selfing offspring generator. It is used to mimic self-fertilization
    in certain plant populations.'''
    def __init__(self, replacement=True, numOffspring = 1, sexMode = RANDOM_SEX,
        ops = SelfingGenoTransmitter(), subPopSize = [], subPops = ALL_AVAIL, weight = 0,
        selectionField = 'fitness'):
        '''Creates a selfing mating scheme where two homologous copies of
        parental chromosomes are transmitted to offspring according to
        Mendelian laws. Please refer to class ``RandomParentChooser`` for
        parameter *replacement* and  *selectionField*, to class
        ``OffspringGenerator`` for parameters *ops*, *sexMode* and
        *numOffspring*, and to class ``HomoMating`` for parameters
        *subPopSize*, *subPops* and *weight*. '''
        HomoMating.__init__(self,
            chooser = RandomParentChooser(replacement, selectionField),
            generator = OffspringGenerator(ops, numOffspring, sexMode),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)

class HermaphroditicMating(HomoMating):
    '''A hermaphroditic mating scheme that chooses two parents randomly
    from the population regardless of sex. The parents could be chosen
    with or without replacement (parameter *replacement*). Selfing (if
    the same parents are chosen) is allowed unless *allowSelfing* is 
    set to *False* '''
    def __init__(self, replacement=True, allowSelfing=True, numOffspring = 1,
        sexMode = RANDOM_SEX, ops = MendelianGenoTransmitter(), 
        subPopSize = [], subPops = ALL_AVAIL, weight = 0,
        selectionField = 'fitness'):
        '''Creates a hermaphroditic mating scheme where individuals can
        serve as father or mother, or both (self-fertilization). Please 
        refer to class ``CombinedParentsChooser`` for parameter *allowSelfing``,
        to ``RandomParentChooser`` for parameter *replacement* and
        *selectionField*, to class ``OffspringGenerator`` for parameters *ops*,
        *sexMode* and *numOffspring*, and to class ``HomoMating`` for parameters
        *subPopSize*, *subPops* and *weight*. '''
        HomoMating.__init__(self,
            chooser = CombinedParentsChooser(
                RandomParentChooser(replacement, selectionField),
                RandomParentChooser(replacement, selectionField),
                allowSelfing=allowSelfing),
            generator = OffspringGenerator(ops, numOffspring, sexMode),
            subPopSize = subPopSize,
            subPops = subPops,
            weight = weight)

## 
## def consanguineousMating(infoFields = [], func = None, param = None,
##         replacement = False, numOffspring = 1.,    sexMode = RANDOM_SEX,
##         ops = MendelianGenoTransmitter(), subPopSize = [],
##         subPops = ALL_AVAIL, weight = 0, selectionField = 'fitness'):
##     '''A homogeneous mating scheme that uses an information parents chooser and
##     a Mendelian offspring generator. A function *func* should be defined to
##     locate certain types of relative to each individual and save their indexes
##     to information fields *infoFields*. This mating scheme will then choose a
##     parent randomly and then another parent from his/her relatives using their
##     saved indexes. Please refer to class ``infoParentsChooser`` for parameters
##     *infoFields*, *func*, *param* and  *selectionField*, to class
##     ``OffspringGenerator`` for parameters *ops*, *sexMode* and *numOffspring*,
##     and to class ``HomoMating`` for parameters *subPopSize*, *subPops* and
##     *weight*.
##     '''
##     return HomoMating(
##         chooser = infoParentsChooser(infoFields, func, param, selectionField),
##         generator = OffspringGenerator(ops, numOffspring, sexMode),
##         subPopSize = subPopSize,
##         subPops = subPops,
##         weight = weight)
## 

class ControlledRandomMating(HomoMating):
    '''A homogeneous mating scheme that uses a random sexual parents chooser
    with replacement and a controlled offspring generator using Mendelian
    genotype transmitter. It falls back to a regular random mating scheme
    if there is no locus to control or no trajectory is defined.'''
    def __init__(self, loci=[], alleles=[], freqFunc=None,
        numOffspring = 1, sexMode = RANDOM_SEX, ops = MendelianGenoTransmitter(),
        subPopSize = [], subPops = ALL_AVAIL, weight = 0, selectionField = 'fitness'):
        '''Creates a random mating scheme that controls allele frequency at
        loci *loci*. At each generation, function *freqFunc* will be called to
        called to obtain intended frequencies of alleles *alleles* at loci
        *loci*. The controlled offspring generator will control the acceptance
        of offspring so that the generation reaches desired allele frequencies
        at these loci. If *loci* is empty or *freqFunc* is ``None``, this mating
        scheme works identically to a ``RandomMating scheme``. Rationals and
        applications of this mating scheme is described in details in a paper *Peng
        et al, 2007 (PLoS Genetics)*. Please refer to class ``RandomParentsChooser``
        for parameters *selectionField*, to class ``ControlledOffspringGenerator``
        for parameters *loci*, *alleles*, *freqFunc*, to class
        ``OffspringGenerator`` for parameters *ops*, *sexMode* and *numOffspring*,
        and to class ``HomoMating`` for parameters *subPopSize*, *subPops* and
        *weight*.
        '''
        if (type(loci) in [type([]), type(())] and len(loci) == 0) or (freqFunc is None):
            HomoMating.__init__(self,
                chooser = RandomParentsChooser(True, selectionField),
                generator = OffspringGenerator(ops, numOffspring, sexMode),
                subPopSize = subPopSize,
                subPops = subPops,
                weight = weight)
        else:
            HomoMating.__init__(self,
                chooser = RandomParentsChooser(True, selectionField),
                generator = ControlledOffspringGenerator(loci, alleles, freqFunc,
                    ops, numOffspring, sexMode),
                subPopSize = subPopSize,
                subPops = subPops,
                weight = weight)


class SNPMutator(MatrixMutator):
    '''A mutator model that assumes two alleles 0 and 1 and accepts mutation
    rate from 0 to 1, and from 1 to 0 alleles. '''
    def __init__(self, u=0, v=0, loci=ALL_AVAIL, mapIn=[], mapOut=[], output='',
        begin=0, end=-1, step=1, at=[], reps=ALL_AVAIL, subPops=ALL_AVAIL,
        infoFields=['ind_id'], lineageMode=FROM_INFO):
        '''Return a ``MatrixMutator`` with proper mutate matrix for a two-allele
        mutation model using mutation rate from allele 0 to 1 (parameter ``u``)
        and from 1 to 0 (parameter ``v``)'''
        MatrixMutator.__init__(self, [[1-u, u], [v, 1-v]], loci, mapIn, mapOut,
            output, begin, end, step, at, reps, subPops, infoFields, lineageMode)


class AcgtMutator(MatrixMutator):
    '''This mutation operator assumes alleles 0, 1, 2, 3 as nucleotides ``A``,
    ``C``, ``G`` and ``T`` and use a 4 by 4 mutation rate matrix to mutate them.
    Although a general model needs 12 parameters, less parameters are needed
    for specific nucleotide mutation models (parameter ``model``). The length
    and meaning of parameter ``rate`` is model dependent.'''
    def __init__(self, rate=[], model='general', loci=ALL_AVAIL, mapIn=[], mapOut=[], output='',
        begin=0, end=-1, step=1, at=[], reps=ALL_AVAIL, subPops=ALL_AVAIL,
        infoFields=['ind_id'], lineageMode=FROM_INFO):
        '''Create a mutation model that mutates between nucleotides ``A``,
        ``C``, ``G``, and ``T`` (alleles are coded in that order as 0, 1, 2
        and 3). Currently supported models are Jukes and Cantor 1969 model
        (``JC69``), Kimura's 2-parameter model (``K80``), Felsenstein 1981
        model (``F81``), Hasgawa, Kishino and Yano 1985 model (``HKY85``),
        Tamura 1992 model (``T92``), Tamura and Nei 1993 model (``TN93``),
        Generalized time reversible model (``GTR``), and a general model
        (``general``) with 12 parameters. Please refer to the simuPOP user's
        guide for detailed information about each model.
        '''
        if model == 'JC69':
            if type(rate) in [type(()), type([])]:
                if len(rate) != 1:
                    raise ValueError('A Jukes and Cantor 1969 model needs one parameter mu.')
                mu = rate[0]
            else:
                mu = rate
            m = [[0,     mu/4., mu/4., mu/4.],
                 [mu/4., 0,     mu/4., mu/4.],
                 [mu/4., mu/4., 0,     mu/4.],
                 [mu/4., mu/4., mu/4., 0    ]]
        elif model == 'K80':
            if len(rate) != 2:
                raise ValueError('A Kimura 2-parameter model requires two parameters mu and k')
            mu, k = rate
            m = [[0,       mu/4.,   mu*k/4., mu/4.  ],
                 [mu/4.,   0,       mu/4.,   mu*k/4.],
                 [mu*k/4., mu/4.,   0,       mu/4.  ],
                 [mu/4.,   mu*k/4., mu/4.,   0      ]]
        elif model == 'F81':
            if len(rate) != 4:
                raise ValueError('A Felsenstein 1981 model requires four parameters mu, pi_A, pi_C and pi_G')
            mu, piA, piC, piG = rate
            piT = 1 - piA - piC - piG
            if piA < 0 or piA > 1 or piC < 0 or piC > 1 or \
                piG < 0 or piG > 1 or piT < 0 or piT > 1:
                raise ValueError('Basic frequencies should be between 0 and 1')
            m = [[0,      mu*piC, mu*piG, mu*piT],
                 [mu*piA, 0,      mu*piG, mu*piT],
                 [mu*piA, mu*piC, 0,      mu*piT],
                 [mu*piA, mu*piC, mu*piG, 0     ]]
        elif model == 'HKY85':
            if len(rate) != 5:
                raise ValueError('A Hasegawa, Kishino and Yano 1985 model requires five parameters mu, k, pi_A, pi_C and pi_G')
            mu, k, piA, piC, piG = rate
            piT = 1 - piA - piC - piG
            if piA < 0 or piA > 1 or piC < 0 or piC > 1 or \
                piG < 0 or piG > 1 or piT < 0 or piT > 1:
                raise ValueError('Basic frequencies should be between 0 and 1')
            m = [[0,        mu*piC,   mu*k*piG, mu*piT  ],
                 [mu*piA,   0,        mu*piG,   mu*k*piT],
                 [mu*k*piA, mu*piC,   0,        mu*piT  ],
                 [mu*piA,   mu*k*piC, mu*piG,   0       ]]
        elif model == 'T92':
            if len(rate) != 2:
                raise ValueError('A Tamura 1992 model requires two parameters mu and pi_GC')
            mu, piGC = rate
            piG = piC = piGC/2.
            piA = piT = (1 - piGC)/2.
            if piA < 0 or piA > 1 or piC < 0 or piC > 1 or \
                piG < 0 or piG > 1 or piT < 0 or piT > 1:
                raise ValueError('Basic frequencies should be between 0 and 1')
            m = [[0,      mu*piC, mu*piG, mu*piT],
                 [mu*piA, 0,      mu*piG, mu*piT],
                 [mu*piA, mu*piC, 0,      mu*piT],
                 [mu*piA, mu*piC, mu*piG, 0     ]]
        elif model == 'TN93':
            if len(rate) != 6:
                raise ValueError('A Tamura and Nei 1993 model requires six parameters mu, k1, k2, pi_A, pi_C and pi_G')
            mu, k1, k2, piA, piC, piG = rate
            piT = 1 - piA - piC - piG
            if piA < 0 or piA > 1 or piC < 0 or piC > 1 or \
                piG < 0 or piG > 1 or piT < 0 or piT > 1:
                raise ValueError('Basic frequencies should be between 0 and 1')
            m = [[0,         mu*piC,    mu*k1*piG, mu*piT   ],
                 [mu*piA,    0,         mu*piG,    mu*k2*piT],
                 [mu*k1*piA, mu*piC,    0,         mu*piT   ],
                 [mu*piA,    mu*k2*piC, mu*piG,    0        ]]
        elif model == 'GTR':
            if len(rate) != 9:
                raise ValueError('A generalized time reversible model requires nine parameters x1, ..., x6, pi_A, pi_C and pi_G')
            x1, x2, x3, x4, x5, x6, piA, piC, piG = rate
            piT = 1 - piA - piC - piG
            if piA < 0 or piA > 1 or piC < 0 or piC > 1 or \
                piG < 0 or piG > 1 or piT < 0 or piT > 1:
                raise ValueError('Basic frequencies should be between 0 and 1')
            m = [[0,  piA*x1/piC, piA*x2/piG, piA*x3/piT],
                 [x1, 0,          piC*x4/piG, piC*x5/piT],
                 [x2, x4,         0,          piG*x6/piT],
                 [x3, x5,         x6,         0         ]]
        elif model == 'general':
            if len(rate) != 12:
                raise ValueError('Please specify 12 parameters for this general nucleotide mutation model')
            m = [[0,       rate[0],  rate[1],  rate[2]],
                 [rate[3], 0,        rate[4],  rate[5]],
                 [rate[6], rate[7],  0,        rate[8]],
                 [rate[9], rate[10], rate[11], 0      ]]
        else:
            raise ValueError('Unrecognized nucleotide mutation model %s' % model)
        MatrixMutator.__init__(self, m, loci, mapIn, mapOut,
            output, begin, end, step, at, reps, subPops, infoFields, lineageMode)


class AminoAcidMutator(MatrixMutator):
    '''
    This operator has not been implemented.
    '''
    def __init__(self, rate=[], model='general', loci=ALL_AVAIL, mapIn=[], mapOut=[], 
        output='', begin=0, end=-1, step=1, at=[], reps=ALL_AVAIL, subPops=ALL_AVAIL,
        infoFields=['ind_id'], lineageMode=FROM_INFO):
        MatrixMutator.__init__(self, rate, loci, mapIn, mapOut,
            output, begin, end, step, at, reps, subPops, infoFields, lineageMode)

#
# functions to corresponding operators
def dump(pop, *args, **kwargs):
    'Apply operator ``Dumper`` to population *pop*.'
    Dumper(*args, **kwargs).apply(pop)

def initSex(pop, *args, **kwargs):
    'Apply operator ``InitSex`` to population *pop*.'
    InitSex(*args, **kwargs).apply(pop)

def initInfo(pop, *args, **kwargs):
    'Apply operator ``InitInfo`` to population *pop*.'
    InitInfo(*args, **kwargs).apply(pop)

def initGenotype(pop, *args, **kwargs):
    'Apply operator ``InitGenotype`` to population *pop*.'
    InitGenotype(*args, **kwargs).apply(pop)

def initLineage(pop, *args, **kwargs):
    'Apply operator ``InitLineage`` to population *pop*.'
    InitLineage(*args, **kwargs).apply(pop)

def pyEval(pop, *args, **kwargs):
    '''Evaluate statements *stmts* (optional) and expression *expr* in
    population *pop*\ 's local namespace and return the result of *expr*.
    If *exposePop* is given, population *pop* will be exposed in its local
    namespace as a variable with a name specified by *exposePop*. Unlike its
    operator counterpart, this function returns the result of *expr* rather
    than writting it to an output.
    '''
    return PyEval(*args, **kwargs).evaluate(pop)

def pyExec(pop, *args, **kwargs):
    '''Execute *stmts* in population *pop*\ 's local namespace.'''
    PyExec(*args, **kwargs).apply(pop)

def infoEval(pop, *args, **kwargs):
    '''Evaluate *expr* for each individual, using information fields as variables.
    Please refer to operator ``InfoEval`` for details.
    '''
    InfoEval(*args, **kwargs).apply(pop)

def infoExec(pop, *args, **kwargs):
    '''Execute *stmts* for each individual, using information fields as variables.
    Please refer to operator ``InfoExec`` for details.
    '''
    InfoExec(*args, **kwargs).apply(pop)

def migrate(pop, *args, **kwargs):
    'Function form of operator ``Migrator``.'
    Migrator(*args, **kwargs).apply(pop)

def backwardMigrate(pop, *args, **kwargs):
    'Function form of operator ``BackwardMigrator``.'
    BackwardMigrator(*args, **kwargs).apply(pop)

def splitSubPops(pop, *args, **kwargs):
    '''Split subpopulations (*subPops*) of population *pop* according to either
    *sizes* or *proportions* of the resulting subpopulations, or an information
    field. Please refer to the operator form of this function (``splitSubPop``)
    for details.'''
    SplitSubPops(*args, **kwargs).apply(pop)

def mergeSubPops(pop, *args, **kwargs):
    '''Merge subpopulations *subPops* of population *pop* into a single
    subpopulation. Please refer to the operator form of this funciton
    (``MergeSubPops``) for details'''
    MergeSubPops(*args, **kwargs).apply(pop)

def resizeSubPops(pop, *args, **kwargs):
    '''Resize subpopulations *subPops* of population *pop* into new sizes
    *size*. Individuals will be added or removed accordingly. Please refer to
    the operator form of this funciton (``ResizeSubPops``) for details'''
    ResizeSubPops(*args, **kwargs).apply(pop)

def matrixMutate(pop, *args, **kwargs):
    'Function form of operator ``MatrixMutator``'
    MatrixMutator(*args, **kwargs).apply(pop)

def snpMutate(pop, *args, **kwargs):
    'Function form of operator ``SNPMutator``'
    SNPMutator(*args, **kwargs).apply(pop)

def acgtMutate(pop, *args, **kwargs):
    'Function form of operator ``AcgtMutator``'
    AcgtMutator(*args, **kwargs).apply(pop)

def kAlleleMutate(pop, *args, **kwargs):
    'Function form of operator ``KAlleleMutator``'
    KAlleleMutator(*args, **kwargs).apply(pop)

def stepwiseMutate(pop, *args, **kwargs):
    'Function form of operator ``StepwiseMutator``'
    StepwiseMutator(*args, **kwargs).apply(pop)

def pyMutate(pop, *args, **kwargs):
    'Function form of operator ``PyMutator``'
    PyMutator(*args, **kwargs).apply(pop)

def mixedMutate(pop, *args, **kwargs):
    'Function form of operator ``MixedMutator``'
    MixedMutator(*args, **kwargs).apply(pop)

def contextMutate(pop, *args, **kwargs):
    'Function form of operator ``ContextMutator``'
    ContextMutator(*args, **kwargs).apply(pop)

def pointMutate(pop, *args, **kwargs):
    'Function form of operator ``PointMutator``'
    PointMutator(*args, **kwargs).apply(pop)

def revertFixedSites(pop, *args, **kwargs):
    'Function form of operator ``RevertFixedSites``'
    RevertFixedSites(*args, **kwargs).apply(pop)

def stat(pop, *args, **kwargs):
    '''Apply operator ``Stat`` with specified parameters to population *pop*.
    Resulting statistics could be accessed from the local namespace of ``pop``
    using functions ``pop.vars()`` or ``pop.dvars()``'''
    Stat(*args, **kwargs).apply(pop)

def tagID(pop, reset=False, *args, **kwargs):
    '''Apply operator ``IdTagger`` to population *pop* to assign a unique ID
    to all individuals in the population. Individuals ID will starts from a
    system wide index. You can reset this start ID using parameter ``reset``
    which can be ``True`` (reset to 1) or a non-negative number (start from
    this number).'''
    if reset != False:
        # True is 1.
        IdTagger().reset(reset)
    IdTagger(*args, **kwargs).apply(pop)

def mapPenetrance(pop, loci, penetrance, ancGens = ALL_AVAIL, *args, **kwargs):
    '''Apply opertor ``MapPenetrance`` to population *pop*. Unlike the
    operator form of this operator that only handles the current generation,
    this function by default assign affection status to all generations.'''
    MapPenetrance(loci, penetrance, ancGens, *args, **kwargs).apply(pop)

def maPenetrance(pop, loci, penetrance, wildtype=0, ancGens = ALL_AVAIL, *args, **kwargs):
    '''Apply opertor ``MaPenetrance`` to population *pop*. Unlike the
    operator form of this operator that only handles the current generation,
    this function by default assign affection status to all generations.'''
    MaPenetrance(loci, penetrance, wildtype, ancGens, *args, **kwargs).apply(pop)

def mlPenetrance(pop, ops, mode, ancGens = ALL_AVAIL, *args, **kwargs):
    '''Apply opertor ``MapPenetrance`` to population *pop*. Unlike the
    operator form of this operator that only handles the current generation,
    this function by default assign affection status to all generations.'''
    MlPenetrance(ops, mode, ancGens, *args, **kwargs).apply(pop)

def pyPenetrance(pop, func, loci=[], ancGens = ALL_AVAIL, *args, **kwargs):
    '''Apply opertor ``PyPenetrance`` to population *pop*. Unlike the
    operator form of this operator that only handles the current generation,
    this function by default assign affection status to all generations.'''
    PyPenetrance(func, loci, ancGens, *args, **kwargs).apply(pop)

def pyMlPenetrance(pop, func, mode, loci=[], ancGens = ALL_AVAIL, *args, **kwargs):
    '''Apply opertor ``PyMlPenetrance`` to population *pop*. Unlike the
    operator form of this operator that only handles the current generation,
    this function by default assign affection status to all generations.'''
    PyMlPenetrance(func, loci, mode, ancGens, *args, **kwargs).apply(pop)

def pyQuanTrait(pop, func, loci=[], ancGens = ALL_AVAIL, *args, **kwargs):
    '''Apply opertor ``PyQuanTrait`` to population *pop*. Unlike the
    operator form of this operator that only handles the current generation,
    this function by default assign affection status to all generations.'''
    PyQuanTrait(func, loci, ancGens, *args, **kwargs).apply(pop)

def discardIf(pop, *args, **kwargs):
    '''Apply operator ``DiscardIf`` to population *pop* to remove individuals according
    to an expression or a Python function.'''
    DiscardIf(*args, **kwargs).apply(pop)

def setRNG(name='', seed=0):
    '''Set random number generator. This function is obsolete but is provided
    for compatibility purposes. Please use setOptions instead'''
    setOptions(name=name, seed=seed)




#!/usr/bin/env python

#
# $File: simuOpt.py $
# $LastChangedDate$
# $Rev$
#
# This file is part of simuPOP, a forward-time population genetics
# simulation environment. Please visit http://simupop.sourceforge.net
# for details.
#
# Copyright (C) 2004 - 2010 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

'''
Module ``simuOpt`` provides a function ``simuOpt.setOptions`` to control which
simuPOP module to load, and how it is loaded, and a class ``simuOpt.Params``
that helps users manage simulation parameters.

When simuPOP is loaded, it checkes a few environmental variables
(``SIMUOPTIMIZED``, ``SIMUALLELETYPE``, and ``SIMUDEBUG``) to determine which
simuPOP module to load, and how to load it. More options can be set using the
``simuOpt.setOptions`` function. For example, you can suppress the banner
message when simuPOP is loaded and require a minimal version of simuPOP for
your script. simuPOP recognize the following commandline arguments

``--optimized``
    Load the optimized version of a simuPOP module.

``--gui=None|batch|interactive|True|wxPython|Tkinter``
    Whether or not use a graphical toolkit and which one to use.
    ``--gui=batch`` is usually used to run a script in batch mode (do not start
    a parameter input dialog and use all default values unless a parameter is
    specified from command line or a configuraiton file. If
    ``--gui=interactive``, an interactive shell will be used to solicit input
    from users. Otherwise, simuPOP will try to use a graphical parameter input
    dialog, and falls to an interactive mode when no graphical Toolkit is
    available. Please refer to parameter ``gui`` for ``simuOpt.setOptions``
    for details.

class ``params.Params`` provides a powerful way to handle commandline
arguments. Briefly speaking, a ``Params`` object can be created from a list
of parameter specification dictionaries. The parameters are then become
attributes of this object. A number of functions are provided to determine
values of these parameters using commandline arguments, a configuration
file, or a parameter input dialog (using ``Tkinter`` or ``wxPython``).
Values of these parameters can be accessed as attributes, or extracted
as a list or a dictionary. Note that the ``Params.getParam`` function
automatically handles the following commandline arguments.

``-h`` or ``--help``
    Print usage message.

``--config=configFile``
    Read parameters from a configuration file *configFile*.

'''


__all__ = [
    'simuOptions',
    'setOptions'
]

import os, sys, re, time, textwrap
#
# simuOptions that will be checked when simuPOP is loaded. This structure
# can be changed by function setOptions
#

simuOptions = {
    'Optimized': False,
    'AlleleType': 'short',
    'Debug': [],
    'Quiet': False,
    'Version': None,
    'Revision': None,
    'GUI': True,
    'Plotter': None,
    'NumThreads': 1,
}

# Optimized: command line option --optimized or environmental variable SIMUOPTIMIZED
if '--optimized' in sys.argv or os.getenv('SIMUOPTIMIZED') is not None:
    simuOptions['Optimized'] = True

# AlleleType: from environmental variable SIMUALLELETYPE
if os.getenv('SIMUALLELETYPE') in ['short', 'long', 'binary', 'mutant', 'lineage']:
   simuOptions['AlleleType'] = os.getenv('SIMUALLELETYPE')
elif os.getenv('SIMUALLELETYPE') is not None:
    print('Environmental variable SIMUALLELETYPE can only be short, long, binary, mutant, or lineage.')

# Debug: from environmental variable SIMUDEBUG
if os.getenv('SIMUDEBUG') is not None:
    simuOptions['Debug'].extend(os.getenv('SIMUDEBUG').split(','))

# openMP number of threads
if os.getenv('OMP_NUM_THREADS') is not None:
    try:
        simuOptions['NumThreads'] = int(os.getenv('OMP_NUM_THREADS'))
    except:
        print('Ignoring invalid value for environmental variable OMP_NUM_THREADS')

# GUI: from environmental variable SIMUGUI
if os.getenv('SIMUGUI') is not None:
    _gui = os.getenv('SIMUGUI')
elif '--gui' in sys.argv:
    if sys.argv[-1] == '--gui':
        raise ValueError('An value is expected for command line option --gui')
    _gui = sys.argv[sys.argv.index('--gui') + 1]
elif True in [x.startswith('--gui=') for x in sys.argv]:
    _gui = sys.argv[[x.startswith('--gui=') for x in sys.argv].index(True)][len('--gui='):]
else:
    _gui = None

if _gui in ['True', 'true', '1']:
    simuOptions['GUI'] = True
elif _gui in ['False', 'false', '0']:
    simuOptions['GUI'] = False
elif _gui in ['wxPython', 'Tkinter', 'batch', 'interactive']:
    simuOptions['GUI'] = _gui
elif _gui is not None:
    print("Invalid value '%s' for environmental variable SIMUGUI or commandline option --gui." % _gui)

def setOptions(alleleType=None, optimized=None, gui=None, quiet=None,
        debug=None, version=None, revision=None, numThreads=None, plotter=None):
    '''Set options before simuPOP is loaded to control which simuPOP module to
    load, and how the module should be loaded.

    alleleType
        Use the standard, binary,long or mutant allele version of the simuPOP
        module if ``alleleType`` is set to 'short', 'binary', 'long', 'mutant',
        or 'lineage' respectively. If this parameter is not set, this function
        will try to get its value from environmental variable ``SIMUALLELETYPE``.
        The standard (short) module will be used if the environmental variable
        is not defined.

    optimized
        Load the optimized version of a module if this parameter is set to
        ``True`` and the standard version if it is set to ``False``. If this
        parameter is not set (``None``), the optimized version will be used
        if environmental variable ``SIMUOPTIMIZED`` is defined. The standard
        version will be used otherwise.
    
    gui
        Whether or not use graphical user interfaces, which graphical toolkit
        to use and how to process parameters in non-GUI mode. If this parameter
        is ``None`` (default), this function will check environmental variable
        ``SIMUGUI`` or commandline option ``--gui`` for a value, and assume
        ``True`` if such an option is unavailable. If ``gui=True``, simuPOP
        will use ``wxPython``-based dialogs if ``wxPython`` is available, and
        use ``Tkinter``-based dialogs if ``Tkinter`` is available and use an
        interactive shell if no graphical toolkit is available.
        ``gui='Tkinter'`` or ``'wxPython'`` can be used to specify the
        graphical toolkit to use. If ``gui='interactive'``, a simuPOP script
        prompt users to input values of parameters. If ``gui='batch'``,
        default values of unspecified parameters will be used. In any case,
        commandline arguments and a configuration file specified by parameter
        --config will be processed. This option is usually left to ``None`` so
        that the same script can be run in both GUI and batch mode using
        commandline option ``--gui``.

    plotter
        (Deprecated)

    quiet
        If set to ``True``, suppress the banner message when a simuPOP module
        is loaded.

    debug
        A list of debug code (as string) that will be turned on when simuPOP
        is loaded. If this parameter is not set, a list of comma separated
        debug code specified in environmental variable ``SIMUDEBUG``, if
        available, will be used. Note that setting ``debug=[]`` will remove
        any debug code that might have been by variable ``SIMUDEBUG``.

    version
        A version string (e.g. 1.0.0) indicating the required version number
        for the simuPOP module to be loaded. simuPOP will fail to load if the
        installed version is older than the required version.

    revision
        Obsolete with the introduction of parameter version.
        
    numThreads
        Number of Threads that will be used to execute a simuPOP script. The
        values can be a positive number (number of threads) or 0 (all available
        cores of the computer, or whatever number set by environmental variable
        ``OMP_NUM_THREADS``). If this parameter is not set, the number of
        threads will be set to 1, or a value set by environmental variable
        ``OMP_NUM_THREADS``.
    '''
    # if the module has already been imported, check which module
    # was imported
    try:
        _imported = sys.modules['simuPOP'].moduleInfo()
    except Exception as e:
        _imported = {}
    # Allele type
    if alleleType in ['long', 'binary', 'short', 'mutant', 'lineage']:
        # if simuPOP has been imported and re-imported with a different module name
        # the existing module will be used so moduleInfo() will return a different
        # module type from what is specified in simuOptions.
        if _imported and _imported['alleleType'] != alleleType:
            raise ImportError(('simuPOP has already been imported with allele type %s (%s) and cannot be '
                're-imported with allele type %s. Please make sure you import module simuOpt before '
                'any simuPOP module is imported.') % (
                    _imported['alleleType'], ('optimized' if _imported['optimized'] else 'standard'),
                    alleleType))
        simuOptions['AlleleType'] = alleleType
    elif alleleType is not None:
        raise TypeError('Parameter alleleType can be either short, long, binary, mutant or lineage.')
    # Optimized
    if optimized in [True, False]:
        # if simuPOP has been imported and re-imported with a different module name
        # the existing module will be used so moduleInfo() will return a different
        # module type from what is specified in simuOptions.
        if _imported and _imported['optimized'] != optimized:
            raise ImportError(('simuPOP has already been imported with allele type %s (%s) and cannot be '
                're-imported in %s mode. Please make sure you import module simuOpt before '
                'any simuPOP module is imported.') % (
                    _imported['alleleType'], ('optimized' if _imported['optimized'] else 'standard'),
                    'optimized' if optimized else 'standard'))
        simuOptions['Optimized'] = optimized
    elif optimized is not None:
        raise TypeError('Parameter optimized can be either True or False.')        
    # Graphical toolkit
    if gui in [True, False, 'wxPython', 'Tkinter', 'batch']:
        simuOptions['GUI'] = gui
    elif gui is not None:
        raise TypeError('Parameter gui can be True/False, wxPython or Tkinter.')
    # Quiet
    if quiet in [True, False]:
        simuOptions['Quiet'] = quiet
    elif quiet is not None:
        raise TypeError('Parameter quiet can be either True or False.')
    # Debug
    if debug is not None:
        if type(debug) == str:
            simuOptions['Debug'] = [debug]
        else:
            simuOptions['Debug'] = debug
    # Version
    if type(version) == str:
        try:
            major, minor, release = [int(x) for x in re.sub('\D', ' ', version).split()]
        except:
            print('Invalid version string %s' % simuOptions['Version'])
        simuOptions['Version'] = version
    elif version is not None:
        raise TypeError('A version string is expected for parameter version.')
    # Revision
    if type(revision) == int:
        simuOptions['Revision'] = revision
    elif revision is not None:
        raise TypeError('A revision number is expected for parameter revision.')
    # NumThreads
    if type(numThreads) == int:
        simuOptions['NumThreads'] = numThreads
    elif numThreads is not None:
        raise TypeError('An integer number is expected for parameter numThreads.')
    if plotter is not None:
        sys.stderr.write('WARNING: plotter option is deprecated because of the removal of rpy/rpy2 support\n')




# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

"""
This module exposes the following GSL (GUN Scientific Library) functions used
by simuPOP to the user interface. Although more functions may be added from time
to time, this module is not intended to become a complete wrapper for GSL. Please
refer to the GSL reference manual (http://www.gnu.org/software/gsl/manual/html_node/)
for details about these functions. Note that random number generation functions
are wrapped into the simuPOP.RNG class.

- ``gsl_cdf_gaussian_P(x, sigma)``
- ``gsl_cdf_gaussian_Q(x, sigma)``
- ``gsl_cdf_gaussian_Pinv(P, sigma)``
- ``gsl_cdf_gaussian_Qinv(Q, sigma)``
- ``gsl_cdf_ugaussian_P(x)``
- ``gsl_cdf_ugaussian_Q(x)``
- ``gsl_cdf_ugaussian_Pinv(P)``
- ``gsl_cdf_ugaussian_Qinv(Q)``
- ``gsl_cdf_exponential_P(x, mu)``
- ``gsl_cdf_exponential_Q(x, mu)``
- ``gsl_cdf_exponential_Pinv(P, mu)``
- ``gsl_cdf_exponential_Qinv(Q, mu)``
- ``gsl_cdf_chisq_P(x, nu)``
- ``gsl_cdf_chisq_Q(x, nu)``
- ``gsl_cdf_chisq_Pinv(P, nu)``
- ``gsl_cdf_chisq_Qinv(Q, nu)``
- ``gsl_cdf_gamma_P(x, a, b)``
- ``gsl_cdf_gamma_Q(x, a, b)``
- ``gsl_cdf_gamma_Pinv(P, a, b)``
- ``gsl_cdf_gamma_Qinv(Q, a, b)``
- ``gsl_ran_gamma_pdf(x, a, b)``
- ``gsl_cdf_beta_P(x, a, b)``
- ``gsl_cdf_beta_Q(x, a, b)``
- ``gsl_cdf_beta_Pinv(P, a, b)``
- ``gsl_cdf_beta_Qinv(Q, a, b)``
- ``gsl_ran_beta_pdf(x, a, b)``
- ``gsl_cdf_binomial_P(k, p, n)``
- ``gsl_cdf_binomial_Q(k, p, n)``
- ``gsl_ran_binomial_pdf(k, p, n)``
- ``gsl_cdf_poisson_P(k, mu)``
- ``gsl_cdf_poisson_Q(k, mu)``
- ``gsl_ran_poisson_pdf(k, mu)``

"""


from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_gsl')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_gsl')
    _gsl = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_gsl', [dirname(__file__)])
        except ImportError:
            import _gsl
            return _gsl
        try:
            _mod = imp.load_module('_gsl', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _gsl = swig_import_helper()
    del swig_import_helper
else:
    import _gsl
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0


def my_error_handler(reason: 'char const *', file: 'char const *', line: 'int', gsl_errno: 'int') -> "void":
    return _gsl.my_error_handler(reason, file, line, gsl_errno)
my_error_handler = _gsl.my_error_handler

def initialize() -> "int":
    return _gsl.initialize()
initialize = _gsl.initialize

def gsl_ran_gamma_pdf(x: 'double const', a: 'double const', b: 'double const') -> "double":
    return _gsl.gsl_ran_gamma_pdf(x, a, b)
gsl_ran_gamma_pdf = _gsl.gsl_ran_gamma_pdf

def gsl_log1p(x: 'double const') -> "double":
    return _gsl.gsl_log1p(x)
gsl_log1p = _gsl.gsl_log1p

def gsl_ran_binomial_pdf(k: 'unsigned int const', p: 'double const', n: 'unsigned int const') -> "double":
    return _gsl.gsl_ran_binomial_pdf(k, p, n)
gsl_ran_binomial_pdf = _gsl.gsl_ran_binomial_pdf

def gsl_ran_beta_pdf(x: 'double const', a: 'double const', b: 'double const') -> "double":
    return _gsl.gsl_ran_beta_pdf(x, a, b)
gsl_ran_beta_pdf = _gsl.gsl_ran_beta_pdf

def gsl_ran_poisson_pdf(k: 'unsigned int const', mu: 'double const') -> "double":
    return _gsl.gsl_ran_poisson_pdf(k, mu)
gsl_ran_poisson_pdf = _gsl.gsl_ran_poisson_pdf

def gsl_cdf_gaussian_P(x: 'double', sigma: 'double') -> "double":
    return _gsl.gsl_cdf_gaussian_P(x, sigma)
gsl_cdf_gaussian_P = _gsl.gsl_cdf_gaussian_P

def gsl_cdf_gaussian_Q(x: 'double', sigma: 'double') -> "double":
    return _gsl.gsl_cdf_gaussian_Q(x, sigma)
gsl_cdf_gaussian_Q = _gsl.gsl_cdf_gaussian_Q

def gsl_cdf_gaussian_Pinv(P: 'double', sigma: 'double') -> "double":
    return _gsl.gsl_cdf_gaussian_Pinv(P, sigma)
gsl_cdf_gaussian_Pinv = _gsl.gsl_cdf_gaussian_Pinv

def gsl_cdf_gaussian_Qinv(Q: 'double', sigma: 'double') -> "double":
    return _gsl.gsl_cdf_gaussian_Qinv(Q, sigma)
gsl_cdf_gaussian_Qinv = _gsl.gsl_cdf_gaussian_Qinv

def gsl_cdf_ugaussian_P(x: 'double') -> "double":
    return _gsl.gsl_cdf_ugaussian_P(x)
gsl_cdf_ugaussian_P = _gsl.gsl_cdf_ugaussian_P

def gsl_cdf_ugaussian_Q(x: 'double') -> "double":
    return _gsl.gsl_cdf_ugaussian_Q(x)
gsl_cdf_ugaussian_Q = _gsl.gsl_cdf_ugaussian_Q

def gsl_cdf_ugaussian_Pinv(P: 'double') -> "double":
    return _gsl.gsl_cdf_ugaussian_Pinv(P)
gsl_cdf_ugaussian_Pinv = _gsl.gsl_cdf_ugaussian_Pinv

def gsl_cdf_ugaussian_Qinv(Q: 'double') -> "double":
    return _gsl.gsl_cdf_ugaussian_Qinv(Q)
gsl_cdf_ugaussian_Qinv = _gsl.gsl_cdf_ugaussian_Qinv

def gsl_cdf_exponential_P(x: 'double', mu: 'double') -> "double":
    return _gsl.gsl_cdf_exponential_P(x, mu)
gsl_cdf_exponential_P = _gsl.gsl_cdf_exponential_P

def gsl_cdf_exponential_Q(x: 'double', mu: 'double') -> "double":
    return _gsl.gsl_cdf_exponential_Q(x, mu)
gsl_cdf_exponential_Q = _gsl.gsl_cdf_exponential_Q

def gsl_cdf_exponential_Pinv(P: 'double', mu: 'double') -> "double":
    return _gsl.gsl_cdf_exponential_Pinv(P, mu)
gsl_cdf_exponential_Pinv = _gsl.gsl_cdf_exponential_Pinv

def gsl_cdf_exponential_Qinv(Q: 'double', mu: 'double') -> "double":
    return _gsl.gsl_cdf_exponential_Qinv(Q, mu)
gsl_cdf_exponential_Qinv = _gsl.gsl_cdf_exponential_Qinv

def gsl_cdf_chisq_P(x: 'double', nu: 'double') -> "double":
    return _gsl.gsl_cdf_chisq_P(x, nu)
gsl_cdf_chisq_P = _gsl.gsl_cdf_chisq_P

def gsl_cdf_chisq_Q(x: 'double', nu: 'double') -> "double":
    return _gsl.gsl_cdf_chisq_Q(x, nu)
gsl_cdf_chisq_Q = _gsl.gsl_cdf_chisq_Q

def gsl_cdf_chisq_Pinv(P: 'double', nu: 'double') -> "double":
    return _gsl.gsl_cdf_chisq_Pinv(P, nu)
gsl_cdf_chisq_Pinv = _gsl.gsl_cdf_chisq_Pinv

def gsl_cdf_chisq_Qinv(Q: 'double', nu: 'double') -> "double":
    return _gsl.gsl_cdf_chisq_Qinv(Q, nu)
gsl_cdf_chisq_Qinv = _gsl.gsl_cdf_chisq_Qinv

def gsl_cdf_gamma_P(x: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_gamma_P(x, a, b)
gsl_cdf_gamma_P = _gsl.gsl_cdf_gamma_P

def gsl_cdf_gamma_Q(x: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_gamma_Q(x, a, b)
gsl_cdf_gamma_Q = _gsl.gsl_cdf_gamma_Q

def gsl_cdf_gamma_Pinv(P: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_gamma_Pinv(P, a, b)
gsl_cdf_gamma_Pinv = _gsl.gsl_cdf_gamma_Pinv

def gsl_cdf_gamma_Qinv(Q: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_gamma_Qinv(Q, a, b)
gsl_cdf_gamma_Qinv = _gsl.gsl_cdf_gamma_Qinv

def gsl_cdf_binomial_P(k: 'unsigned int', p: 'double', n: 'unsigned int') -> "double":
    return _gsl.gsl_cdf_binomial_P(k, p, n)
gsl_cdf_binomial_P = _gsl.gsl_cdf_binomial_P

def gsl_cdf_binomial_Q(k: 'unsigned int', p: 'double', n: 'unsigned int') -> "double":
    return _gsl.gsl_cdf_binomial_Q(k, p, n)
gsl_cdf_binomial_Q = _gsl.gsl_cdf_binomial_Q

def gsl_cdf_beta_P(x: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_beta_P(x, a, b)
gsl_cdf_beta_P = _gsl.gsl_cdf_beta_P

def gsl_cdf_beta_Q(x: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_beta_Q(x, a, b)
gsl_cdf_beta_Q = _gsl.gsl_cdf_beta_Q

def gsl_cdf_beta_Pinv(P: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_beta_Pinv(P, a, b)
gsl_cdf_beta_Pinv = _gsl.gsl_cdf_beta_Pinv

def gsl_cdf_beta_Qinv(Q: 'double', a: 'double', b: 'double') -> "double":
    return _gsl.gsl_cdf_beta_Qinv(Q, a, b)
gsl_cdf_beta_Qinv = _gsl.gsl_cdf_beta_Qinv

def gsl_cdf_poisson_P(k: 'unsigned int const', mu: 'double const') -> "double":
    return _gsl.gsl_cdf_poisson_P(k, mu)
gsl_cdf_poisson_P = _gsl.gsl_cdf_poisson_P

def gsl_cdf_poisson_Q(k: 'unsigned int const', mu: 'double const') -> "double":
    return _gsl.gsl_cdf_poisson_Q(k, mu)
gsl_cdf_poisson_Q = _gsl.gsl_cdf_poisson_Q
# This file is compatible with both classic and new-style classes.





#!/usr/bin/env python

#
# $File: sampling.py $
# $LastChangedDate: 2009-10-11 17:41:30 -0500 (Sun, 11 Oct 2009) $
# $Rev: 3033 $
#
# This file is part of simuPOP, a forward-time population genetics
# simulation environment. Please visit http://simupop.sourceforge.net
# for details.
#
# Copyright (C) 2004 - 2009 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#


"""
This module provides classes and functions that could be used to draw samples
from a simuPOP population. These functions accept a list of parameters such
as ``subPops`` ((virtual) subpopulations from which samples will be drawn) and
``numOfSamples`` (number of samples to draw) and return a list of populations. Both
independent individuals and dependent individuals (Pedigrees) are supported.

Independent individuals could be drawn from any Population. pedigree
information is not necessary and is usually ignored. Unique IDs are not needed
either although such IDs could help you identify samples in the parent
Population.

Pedigrees could be drawn from multi-generational populations or age-structured
populations. All individuals are required to have a unique ID (usually tracked
by operator ``IdTagger`` and are stored in information field ``ind_id``).
Parents of individuals are usually tracked by operator ``PedigreeTagger`` and
are stored in information fields ``father_id`` and ``mother_id``. If parental
information is tracked using operator ``ParentsTagger`` and information fields
``father_idx`` and ``mother_idx``, a function ``sampling.indexToID`` can be
used to convert index based pedigree to ID based Pedigree. Note that
``ParentsTagger`` can not be used to track Pedigrees in age-structured
populations because they require parents of each individual resides in a
parental generation.

All sampling functions support virtual subpopulations through parameter
``subPops``, although sample size specification might vary. This feature
allows you to draw samples with specified properties. For example, you
could select only female individuals for cases of a female-only disease,
or select individuals within certain age-range. If you specify a list
of (virtual) subpopulations, you are usually allowed to draw certain
number of individuals from each subpopulation.
"""

__all__ = [
    #
    'indexToID',
    # Classes that can be derived to implement more complicated
    # sampling scheme
    'BaseSampler',
    'RandomSampler',
    'CaseControlSampler',
    'PedigreeSampler',
    'AffectedSibpairSampler',
    'NuclearFamilySampler',
    'ThreeGenFamilySampler',
    'CombinedSampler',
    # Functions to draw samples
    'drawRandomSample',
    'drawRandomSamples',
    'drawCaseControlSample',
    'drawCaseControlSamples',
    'drawAffectedSibpairSample',
    'drawAffectedSibpairSamples',
    'drawNuclearFamilySample',
    'drawNuclearFamilySamples',
    'drawThreeGenFamilySample',
    'drawThreeGenFamilySamples',
    'drawCombinedSample',
    'drawCombinedSamples',
    #
]

from simuPOP import ALL_AVAIL, Pedigree, OUTBRED_SPOUSE, COMMON_OFFSPRING, FEMALE_ONLY, \
    MALE, AFFECTED, tagID, getRNG

import random
def random_shuffle(x):
    random.shuffle(x, getRNG().randUniform)

def isSequence(obj):
    return hasattr(obj, '__iter__')

def isNumber(obj):
    return isinstance(obj, (int, float))

def indexToID(pop, idField='ind_id', fatherField='father_id', motherField='mother_id',
              fatherIndex='father_idx', motherIndex='mother_idx', reset=False):
    '''This function adds information field idField (default to ``'ind_id'``)
    to population ``pop`` and assigns an unique ID to every individual in this
    Population. It then adds information fields fatherField (default to
    ``'fatherField'``) and motherField (default to ``'motherField'``) and set
    their values with IDs according to the established index based
    parents-children relationship. Existing information fields will be used if
    idField, fatherField or motherField already exist. This function uses a
    system-wide ID generator for unique IDs, which does not have to start from
    zero. A parameter ``reset`` could be used to reset starting ID to zero
    (if ``reset=True``) or a specified number (``reset=number``).
    '''
    pop.addInfoFields([idField, fatherField, motherField], -1)
    # set each individual's unique ID to idField
    tagID(pop, reset=reset, infoFields=idField)
    # save each individual's parents' IDs to fatherField and motherField
    for gen in range(pop.ancestralGens()-1, -1, -1):
        pop.useAncestralGen(gen)
        for ind in pop.individuals():
            if ind.info(fatherIndex) != -1:
                father = pop.ancestor(ind.info(fatherIndex), gen+1)
                ind.setInfo(father.info(idField), fatherField)
            if ind.info(motherIndex) != -1:
                mother = pop.ancestor(ind.info(motherIndex), gen+1)
                ind.setInfo(mother.info(idField), motherField)


# Sampling classes and functions

class BaseSampler:
    '''
    A sampler extracts individuals from a simuPOP population and return them
    as separate populations. This base class defines the common interface of
    all sampling classes, including how samples prepared and returned.
    '''
    def __init__(self, subPops=ALL_AVAIL):
        '''Create a sampler with parameter ``subPops``, which will be used
        to prepare population for sampling. ``subPops`` should be a list of
        (virtual) subpopulations from which samples are drawn. The default
        value is ALL_AVAIL, which means all available subpopulations of a
        Population.
        '''
        self.subPops=subPops
        self.pop = None

    def prepareSample(self, pop, rearrange):
        '''Prepare passed population object for sampling according to parameter
        ``subPops``. If samples are drawn from the whole population, a
        Population will be trimmed if only selected (virtual) subpopulations
        are used. If samples are drawn separately from specified subpopulations,
        Population ``pop`` will be rearranged (if ``rearrange==True``) so that
        each subpoulation corresponds to one element in parameter ``subPops``.
        '''
        if self.subPops == ALL_AVAIL:
            self.pop = pop
        else:
            self.pop = pop.extractSubPops(self.subPops, rearrange)
        return True

    def drawSample(self, pop):
        '''
        Draw and return a sample.
        '''
        raise SystemError('Please re-implement this drawSample function in the derived class.')

    def drawSamples(self, pop, numOfSamples):
        '''
        Draw multiple samples and return a list of populations.
        '''
        if numOfSamples < 0:
            raise ValueError("Negative number of samples are unacceptable")
        # 
        return [self.drawSample(pop) for x in range(numOfSamples)]


class RandomSampler(BaseSampler):
    '''A sampler that draws individuals randomly.
    '''
    def __init__(self, sizes, subPops=ALL_AVAIL):
        '''Creates a random sampler with specified number of individuals.
        '''
        BaseSampler.__init__(self, subPops)
        self.sizes = sizes

    def drawSample(self, input_pop):
        '''Draw a random sample from passed population.
        '''
        if self.pop is None:
            # this will produce self.pop.
            self.prepareSample(input_pop, isSequence(self.sizes))
        #
        if not isSequence(self.sizes):
            size = self.sizes
            if size > self.pop.popSize():
                print('Warning: sample size %d is greater than population size %d.' % (size, self.pop.popSize()))
                size = self.pop.popSize()
            # randomly choose size individuals
            values = list(range(self.pop.popSize()))
            random_shuffle(values)
            indexes = values[:size]
        else:
            indexes = []
            for sp in range(self.pop.numSubPop()):
                size = self.sizes[sp]
                if size > self.pop.subPopSize(sp):
                    print('Warning: sample size (%d) at subpopulation %d is greater than subpopulation size %d ' \
                        % (size, sp, self.pop.subPopSize(sp)))
                values = list(range(self.pop.subPopBegin(sp), self.pop.subPopEnd(sp)))
                random_shuffle(values)
                indexes.extend(values[:size])
        return self.pop.extractIndividuals(indexes = indexes)


def drawRandomSample(pop, sizes, subPops=ALL_AVAIL):
    '''Draw ``sizes`` random individuals from a population. If a single ``sizes``
    is given, individuals are drawn randomly from the whole population or
    from specified (virtual) subpopulations (parameter ``subPops``). Otherwise,
    a list of numbers should be used to specify number of samples from each
    subpopulation, which can be all subpopulations if ``subPops=ALL_AVAIL``
    (default), or from each of the specified (virtual) subpopulations. This
    function returns a population with all extracted individuals.
    '''
    return RandomSampler(sizes=sizes, subPops=subPops).drawSample(pop)


def drawRandomSamples(pop, sizes, numOfSamples=1, subPops=ALL_AVAIL):
    '''Draw ``numOfSamples`` random samples from a population and return a list of
    populations. Please refer to function ``drawRandomSample`` for more details
    about parameters ``sizes`` and ``subPops``.'''
    return RandomSampler(sizes=sizes, subPops=subPops).drawSamples(pop, numOfSamples=numOfSamples)


class CaseControlSampler(BaseSampler):
    '''A sampler that draws affected and unaffected individuals randomly.
    '''
    def __init__(self, cases, controls, subPops=ALL_AVAIL):
        '''Ceates a case-control sampler with specified number of cases
        and controls.
        '''
        BaseSampler.__init__(self, subPops)
        self.cases = cases
        self.controls = controls
        if type(self.cases) != type(controls):
            raise ValueError("Parameter cases and controls should have the same type.")
        if isSequence(self.cases) and isSequence(self.controls) and \
            len(self.cases) != len(self.controls):
            raise ValueError("Cases and controls should have the same type")

    def prepareSample(self, input_pop):
        '''Find out indexes all affected and unaffected individuales.
        '''
        BaseSampler.prepareSample(self, input_pop, isSequence(self.cases))
        if self.pop is None:
            # this will produce self.pop and self.cases and self.controls
            self.prepareSample(input_pop)
        #
        if not isSequence(self.cases):
            # find affected individuals
            self.affected = []
            self.unaffected = []
            for idx,ind in enumerate(self.pop.individuals()):
                if ind.affected():
                    self.affected.append(idx)
                else:
                    self.unaffected.append(idx)
            #
            if self.cases > len(self.affected):
                print('Warning: number of cases %d is greater than number of affectedected individuals %d.' \
                    % (self.cases, len(self.affected)))
            #
            if self.controls > len(self.unaffected):
                print('Warning: number of controls %d is greater than number of affectedected individuals %d.' \
                    % (self.controls, len(self.unaffected)))
        else:
            if len(self.cases) != self.pop.numSubPop():
                raise ValueError('If an list of cases is given, it should be specified for all subpopulations')
            self.affected = []
            self.unaffected = []
            for sp in range(self.pop.numSubPop()):
                # find self.affectedected individuals
                aff = []
                unaff = []
                for idx in range(self.pop.subPopBegin(sp), self.pop.subPopEnd(sp)):
                    if self.pop.individual(idx).affected():
                        aff.append(idx)
                    else:
                        unaff.append(idx)
                #
                if self.cases[sp] > len(aff):
                    print('Warning: number of cases %d is greater than number of affectedected individuals %d in subpopulation %d.' \
                        % (self.cases[sp], len(aff), sp))
                #
                if self.controls[sp] > len(unaff):
                    print('Warning: number of controls %d is greater than number of affectedected individuals %d in subpopulation %d.' \
                        % (self.controls[sp], len(unaff), sp))
                self.affected.append(aff)
                self.unaffected.append(unaff)

    def drawSample(self, input_pop):
        '''Draw a case control sample
        '''
        if self.pop is None:
            # this will produce self.pop and self.affected and self.unaffected
            self.prepareSample(input_pop)
        #
        if not isSequence(self.cases):
            random_shuffle(self.affected)
            random_shuffle(self.unaffected)
            indexes = self.affected[:self.cases] + self.unaffected[:self.controls]
        else:
            indexes = []
            for sp in range(self.pop.numSubPop()):
                random_shuffle(self.affected[sp])
                random_shuffle(self.unaffected[sp])
                indexes.extend(self.affected[sp][:self.cases[sp]])
                indexes.extend(self.unaffected[sp][:self.controls[sp]])
        return self.pop.extractIndividuals(indexes = indexes)


def drawCaseControlSample(pop, cases, controls, subPops=ALL_AVAIL):
    '''Draw a case-control samples from a population with ``cases``
    affected and ``controls`` unaffected individuals. If single ``cases`` and
    ``controls`` are given, individuals are drawn randomly from the whole
    Population or from specified (virtual) subpopulations (parameter
    ``subPops``). Otherwise, a list of numbers should be used to specify
    number of cases and controls from each subpopulation, which can be all
    subpopulations if ``subPops=ALL_AVAIL`` (default), or from each of the
    specified (virtual) subpopulations. This function returns a population with
    all extracted individuals.
    '''
    return CaseControlSampler(cases, controls, subPops).drawSample(pop) 


def drawCaseControlSamples(pop, cases, controls, numOfSamples=1, subPops=ALL_AVAIL):
    '''Draw ``numOfSamples`` case-control samples from a population with ``cases``
    affected and ``controls`` unaffected individuals and return a list of
    populations. Please refer to function ``drawCaseControlSample`` for a
    detailed descriptions of parameters.
    '''
    return CaseControlSampler(cases, controls, subPops).drawSamples(pop, numOfSamples) 


class PedigreeSampler(BaseSampler):
    '''The base class of all pedigree based sampler.
    '''
    def __init__(self, families, subPops=ALL_AVAIL, idField='ind_id',
        fatherField='father_id', motherField='mother_id'):
        '''Creates a pedigree sampler with parameters

        families
            number of families. This can be a number or a list of numbers. In
            the latter case, specified families are drawn from each
            subpopulation.

        subPops
            A list of (virtual) subpopulations from which samples are drawn.
            The default value is ALL_AVAIL, which means all available
            subpopulations of a population.
        '''
        BaseSampler.__init__(self, subPops)
        self.families = families
        self.idField = idField
        self.fatherField = fatherField
        self.motherField = motherField
        self.pedigree = None

    def prepareSample(self, pop, loci=[], infoFields=[], ancGens=ALL_AVAIL):
        '''
        Prepare self.pedigree, some pedigree sampler might need additional loci and
        information fields for this sampler.
        '''
        # create self.pop
        BaseSampler.prepareSample(self, pop, isSequence(self.families))
        # get self.pedigree
        self.pedigree = Pedigree(self.pop, loci, infoFields,
            ancGens, self.idField, self.fatherField, self.motherField)
        # Your prepareSample should define selected_IDs, which should be the
        # anchor individuals of each Pedigree, typically father or grandfather.
        # Other family members will be looked up through function self.family.
        self.selected_IDs = []

    def family(self, id):
        'Get the family of individual with id.'
        return [id]

    def drawSample(self, input_pop):
        'Randomly select Pedigrees'
        if self.pedigree is None:
            # this will give us self.pop, self.pedigree, and self.selectedIDs
            self.prepareSample(input_pop)
        #
        if not isSequence(self.families):
            if self.families > len(self.selectedIDs):
                print('Warning: number of requested Pedigrees %d is greater than what exists (%d).' \
                    % (self.families, len(self.selectedIDs)))
            # a tuple might be returned
            self.selectedIDs = list(self.selectedIDs)
            random_shuffle(self.selectedIDs)
            # we select families one by one to exclude overlapping families.
            IDs = set()
            cnt = 0
            for id in self.selectedIDs:
                fam = set(self.family(id))
                if len(IDs & fam) == 0:
                    IDs.update(fam)
                    cnt += 1
                if cnt == self.families:
                    break
            if cnt != self.families:
                print('Warning: not enough non-overlapping Pedigrees are found (requested %d, found %d).' \
                    % (self.families, cnt))
        else:
            IDs = set()
            for sp in range(self.pop.numSubPop()):
                if self.families[sp] > len(self.selectedIDs[sp]):
                    print('Warning: number of requested Pedigrees %d is greater than what exists (%d) in subpopulation %d.' \
                        % (self.families[sp], len(self.selectedIDs[sp]), sp))
                #
                # a tuple might be returned
                self.selectedIDs[sp] = list(self.selectedIDs[sp])
                random_shuffle(self.selectedIDs[sp])
                # we select families one by one to exclude overlapping families.
                cnt = 0
                for id in self.selectedIDs[sp]:
                    fam = set(self.family(id))
                    if len(IDs & fam) == 0:
                        IDs.update(fam)
                        cnt += 1
                    if cnt == self.families[sp]:
                        break
                if cnt != self.families[sp]:
                    print('Warning: not enough non-overlapping Pedigrees are found (requested %d, found %d).' \
                        % (self.families, cnt))
        # get family members
        return self.pop.extractIndividuals(IDs = list(IDs), idField = self.idField)


class AffectedSibpairSampler(PedigreeSampler):
    '''A sampler that draws a nuclear family with two affected offspring.
    '''
    def __init__(self, families, subPops=ALL_AVAIL, idField='ind_id',
        fatherField='father_id', motherField='mother_id'):
        '''Initialize an affected sibpair sampler.'''
        PedigreeSampler.__init__(self, families, subPops, idField, fatherField, motherField)

    def family(self, id):
        'Return id, its spouse and their children'
        ind = self.pedigree.indByID(id)
        return id, ind.spouse, ind.off1, ind.off2

    def prepareSample(self, input_pop):
        'Find the father or all affected sibpair families'
        # this will give us self.pop and self.pedigree
        PedigreeSampler.prepareSample(self, input_pop, isSequence(self.families))
        if isSequence(self.families) and len(self.families) != self.pop.numSubPop():
            raise ValueError('If an list of family counts is given, it should be specified for all subpopulations')
        #
        # locate all affected siblings
        self.pedigree.addInfoFields(['spouse', 'off1', 'off2'])
        # only look for wife so families will not overlap
        self.pedigree.locateRelatives(OUTBRED_SPOUSE, ['spouse'], FEMALE_ONLY)
        # look for affected offspring
        self.pedigree.locateRelatives(COMMON_OFFSPRING, ['spouse', 'off1', 'off2'], affectionStatus=AFFECTED)
        # find qualified families
        if not isSequence(self.families):
            self.selectedIDs = self.pedigree.individualsWithRelatives(['spouse', 'off1', 'off2'])
        else:
            self.selectedIDs = []
            for sp in range(self.pedigree.numSubPop()):
                self.selectedIDs.append(self.pedigree.individualsWithRelatives(['spouse', 'off1', 'off2'],
                    subPops=sp))


def drawAffectedSibpairSample(pop, families, subPops=ALL_AVAIL, 
    idField='ind_id', fatherField='father_id', motherField='mother_id'):
    '''Draw affected sibpair samples from a population. If a single
    ``families`` is given, affected sibpairs and their parents are drawn
    randomly from the whole population or from specified (virtual)
    subpopulations (parameter ``subPops``). Otherwise, a list of numbers should
    be used to specify number of families from each subpopulation, which can be
    all subpopulations if ``subPops=ALL_AVAIL`` (default), or from each of the
    specified (virtual) subpopulations. This function returns a population that
    contains extracted individuals.
    '''
    return AffectedSibpairSampler(families, subPops, idField, fatherField,
        motherField).drawSample(pop)
 

def drawAffectedSibpairSamples(pop, families, numOfSamples=1, subPops=ALL_AVAIL, 
    idField='ind_id', fatherField='father_id', motherField='mother_id'):
    '''Draw ``numOfSamples`` affected sibpair samplesa from population ``pop`` and
    return a list of populations. Please refer to function
    ``drawAffectedSibpairSample`` for a description of other parameters.
    '''
    return affectedSibpairSample(families, subPops, idField, fatherField,
        motherField).drawSamples(pop, numOfSamples)


class NuclearFamilySampler(PedigreeSampler):
    '''A sampler that draws nuclear families with specified number of affected
    parents and offspring.
    '''
    def __init__(self, families, numOffspring, affectedParents=0, affectedOffspring=0,
        subPops=ALL_AVAIL, idField='ind_id', fatherField='father_id', motherField='mother_id'):
        '''Creates a nuclear family sampler with parameters

        families
            number of families. This can be a number or a list of numbers. In the latter
            case, specified families are drawn from each subpopulation.

        numOffspring
            number of offspring. This can be a fixed number or a range [min, max].

        affectedParents
            number of affected parents. This can be a fixed number or a range [min, max].

        affectedOffspring
            number of affected offspring. This can be a fixed number of a range [min, max].

        subPops
            A list of (virtual) subpopulations from which samples are drawn.
            The default value is ALL_AVAIL, which means all available
            subpopulations of a population.
        '''
        if isNumber(numOffspring):
            if numOffspring < 1:
                raise ValueError('Number of offsprings must be equal to or larger than 1.')
            self.numOffspring = numOffspring, numOffspring
        elif isSequence(numOffspring):
            if len(numOffspring) != 2:
                raise ValueError('Two boundary numbers are needed for the allowed range of number of offsprings')
            if numOffspring[0] < 1 or numOffspring[0] > numOffspring[1]:
                raise ValueError('Minimum number of offsprings must not be smaller than 1 or larger than maximum number of offsprings.')
            self.numOffspring = numOffspring
        else:
            raise ValueError('Number of offsprings should be an integer number or a range of allowed values.')
        #
        if isNumber(affectedParents):
            if affectedParents not in [0, 1, 2]:
                raise ValueError('Number of affected individuals in parents can only take 0 or 1 or 2.')
            self.affectedParents = affectedParents, affectedParents
        elif isSequence(affectedParents):
            if len(affectedParents) != 2:
                raise ValueError('Two boundary numbers are needed for the range of number of affected parents.')
            if affectedParents[0] not in [0,1,2] or affectedParents[1] not in [0,1,2] or affectedParents[0] > affectedParents[1]:
                raise ValueError('Range of number of affected parents must be within (0, 2).')
            self.affectedParents = affectedParents
        else:
            raise ValueError('Number of affected parents should be an integer number (<= 2) or a range of allowed values.')
        #
        if isNumber(affectedOffspring):
            if affectedOffspring > self.numOffspring[1]:
                raise ValueError('Number of affected offsprings cannot be larger than number of offsprings.')
            self.affectedOffspring = affectedOffspring, affectedOffspring
        elif isSequence(affectedOffspring):
            if len(affectedOffspring) != 2:
                raise ValueError('Two boundary numbers are needed for the range of number of affected offsprings.')
            if affectedOffspring[0] > self.numOffspring[1]:
                raise ValueError('Minimum number of affected offsprings cannot be larger than number of offsprings.')
            self.affectedOffspring = affectedOffspring
        else:
            raise ValueError('Number of affected offsprings should be a proper integer nubmer or a range of allowed values.')
        #
        PedigreeSampler.__init__(self, families, subPops, idField, fatherField, motherField)

    def family(self, id):
        'Return id, its spouse and their children'
        ind = self.pedigree.indByID(id)
        offIDs = [ind.info('off%d' % x) for x in range(self.numOffspring[1])]
        return [id, ind.spouse] + [x for x in offIDs if x >= 0]

    def prepareSample(self, input_pop):
        # this will give us self.pop and self.pedigree
        PedigreeSampler.prepareSample(self, input_pop, isSequence(self.families))
        if isSequence(self.families) and len(self.families) != self.pop.numSubPop():
            raise ValueError('If an list of family counts is given, it should be specified for all subpopulations')
        #
        # locate all affected siblings
        minOffFields = ['off%d' % x for x in range(self.numOffspring[0])]
        offFields = ['off%d' % x for x in range(self.numOffspring[1])]
        self.pedigree.addInfoFields(['spouse'] + offFields)
        # only look for wife so families will not overlap
        self.pedigree.locateRelatives(OUTBRED_SPOUSE, ['spouse'], FEMALE_ONLY)
        # look for offspring
        self.pedigree.locateRelatives(COMMON_OFFSPRING, ['spouse'] + offFields)
        # check number of affected individuals and filter them out.
        def qualify(id):
            father = self.pedigree.indByID(id)
            mother = self.pedigree.indByID(father.info('spouse'))
            parAff = father.affected() + mother.affected()
            if parAff < self.affectedParents[0] or parAff > self.affectedParents[1]:
                return False
            offID = [father.info('off%d' % x) for x in range(self.numOffspring[1])]
            offAff = sum([self.pedigree.indByID(id).affected() for id in offID if id >= 0])
            if offAff < self.affectedOffspring[0] or offAff > self.affectedOffspring[1]:
                return False
            return True
        # find all families with at least minOffFields...
        if not isSequence(self.families):
            self.selectedIDs = list(filter(qualify, self.pedigree.individualsWithRelatives(['spouse'] + minOffFields)))
        else:
            self.selectedIDs = []
            for sp in range(self.pedigree.numSubPop()):
                self.selectedIDs.append(list(filter(qualify, self.pedigree.individualsWithRelatives(['spouse'] + minOffFields, subPops=sp))))


def drawNuclearFamilySample(pop, families, numOffspring, affectedParents=0,
    affectedOffspring=0, subPops=ALL_AVAIL, idField='ind_id', fatherField='father_id',
    motherField='mother_id'):
    '''Draw nuclear families from a population. Number of offspring, number of
    affected parents and number of affected offspring should be specified using
    parameters ``numOffspring``, ``affectedParents`` and ``affectedOffspring``,
    which can all be a single number, or a range ``[a, b]`` (``b`` is incldued).
    If a single ``families`` is given, Pedigrees are drawn randomly from the
    whole population or from specified (virtual) subpopulations (parameter
    ``subPops``). Otherwise, a list of numbers should be used to specify
    numbers of families from each subpopulation, which can be all
    subpopulations if ``subPops=ALL_AVAIL`` (default), or from each of the
    specified (virtual) subpopulations. This function returns a population that
    contains extracted individuals.
    '''
    return NuclearFamilySampler(families, numOffspring, affectedParents,
        affectedOffspring, subPops, idField, fatherField, motherField).drawSample(pop)
 

def drawNuclearFamilySamples(pop, families, numOffspring, affectedParents=0,
    affectedOffspring=0, numOfSamples=1, subPops=ALL_AVAIL, idField='ind_id',
    fatherField='father_id', motherField='mother_id'):
    '''Draw ``numOfSamples`` affected sibpair samplesa from population ``pop`` and
    return a list of populations. Please refer to function
    ``drawNuclearFamilySample`` for a description of other parameters.
    '''
    return nuclearFamilySample(families, numOffspring, affectedParents,
        affectedOffspring, subPops, idField, fatherField,
        motherField).drawSamples(pop, numOfSamples)


class ThreeGenFamilySampler(PedigreeSampler):
    '''A sampler that draws three-generation families with specified pedigree
    size and number of affected individuals.
    '''
    def __init__(self, families, numOffspring, pedSize, numOfAffected=0,
        subPops=ALL_AVAIL, idField='ind_id', fatherField='father_id', motherField='mother_id'):
        '''
        families
            number of families. This can be a number or a list of numbers. In the latter
            case, specified families are drawn from each subpopulation.

        numOffspring
            number of offspring. This can be a fixed number or a range [min, max].

        pedSize
            number of individuals in the Pedigree. This can be a fixed number or
            a range [min, max].

        numAfffected
            number of affected individuals in the Pedigree. This can be a fixed number
            or a range [min, max]

        subPops
            A list of (virtual) subpopulations from which samples are drawn.
            The default value is ALL_AVAIL, which means all available
            subpopulations of a population.
        '''
        if isNumber(numOffspring):
            if numOffspring < 1:
                raise ValueError('Number of offsprings must be equal to or larger than 1.')
            self.numOffspring = numOffspring, numOffspring
        elif isSequence(numOffspring):
            if len(numOffspring) != 2:
                raise ValueError('Two boundary numbers are needed for the allowed range of number of offsprings')
            if numOffspring[0] < 1 or numOffspring[0] > numOffspring[1]:
                raise ValueError('Minimum number of offsprings must not be smaller than 1 or larger than maximum number of offsprings.')
            self.numOffspring = numOffspring
        else:
            raise ValueError('Number of offsprings should be an integer number or a range of allowed values.')
        #
        if isNumber(pedSize):
            self.pedSize = pedSize, pedSize
        elif isSequence(pedSize):
            if len(pedSize) != 2:
                raise ValueError('Two boundary numbers are needed for the range of number of individuals in a Pedigree.')
            self.pedSize = pedSize
        else:
            raise ValueError('Number of affected parents should be an integer number (<= 1) or a range of allowed values.')
        #
        if isNumber(numOfAffected):
            if numOfAffected > self.pedSize[1]:
                raise ValueError('Number of affected individuals cannot be larger than pedigree size.')
            self.numOfAffected = numOfAffected, numOfAffected
        elif isSequence(numOfAffected):
            if len(numOfAffected) != 2:
                raise ValueError('Two boundary numbers are needed for the range of number of affected individuals.')
            if numOfAffected[0] > self.pedSize[1]:
                raise ValueError('Minimum number of affected offsprings cannot be larger than number of individuals in a Pedigree.')
            self.numOfAffected = numOfAffected
        else:
            raise ValueError('Number of affected offsprings should be a proper integer nubmer or a range of allowed values.')
        #
        PedigreeSampler.__init__(self, families, subPops, idField, fatherField, motherField)

    def family(self, id):
        '''Return id, its spouse, their children, children's spouse and grandchildren'''
        father = self.pedigree.indByID(id)
        spouseID = father.spouse
        offID = [father.info('off%d' % x) for x in range(self.numOffspring[1])]
        offID = [x for x in offID if x >= 1]
        offSpouseID = [self.pedigree.indByID(x).spouse for x in offID]
        offSpouseID = [x for x in offSpouseID if x >= 1]
        grandOffID = [father.info('goff%d' % x) for x in range(self.numOffspring[1]**2)]
        grandOffID = [x for x in grandOffID if x >= 1]
        return [id, spouseID] + offID + offSpouseID + grandOffID

    def prepareSample(self, input_pop):
        # this will give us self.pop and self.pedigree
        PedigreeSampler.prepareSample(self, input_pop, isSequence(self.families))
        if isSequence(self.families) and len(self.families) != self.pop.numSubPop():
            raise ValueError('If an list of family counts is given, it should be specified for all subpopulations')
        #
        # locate all affected siblings
        minOffFields = ['off%d' % x for x in range(self.numOffspring[0])]
        offFields = ['off%d' % x for x in range(self.numOffspring[1])]
        minGrandOffFields = ['goff%d' % x for x in range(self.numOffspring[0]**2)]
        grandOffFields = ['goff%d' % x for x in range(self.numOffspring[1]**2)]
        self.pedigree.addInfoFields(['spouse'] + offFields + grandOffFields)
        # only look for wife so families will not overlap
        self.pedigree.locateRelatives(OUTBRED_SPOUSE, ['spouse'], FEMALE_ONLY)
        # look for offspring
        self.pedigree.locateRelatives(COMMON_OFFSPRING, ['spouse'] + offFields)
        # look for grand children
        self.pedigree.traceRelatives(fieldPath = [offFields, offFields], resultFields = grandOffFields)
        # check number of affected individuals and filter them out.
        def qualify(id):
            father = self.pedigree.indByID(id)
            mother = self.pedigree.indByID(father.spouse)
            offID = [father.info('off%d' % x) for x in range(self.numOffspring[1])]
            offID = [x for x in offID if x >= 0]
            offSpouseID = [self.pedigree.indByID(id).spouse for x in offID]
            grandOffID = [father.info('goff%d' % x) for x in range(self.numOffspring[1]**2)]
            grandOffID = [x for x in grandOffID if x >= 0]
            pedSize = 2 + len(offID) + len(offSpouseID) + len(grandOffID)
            if pedSize < self.pedSize[0] or pedSize > self.pedSize[1]:
                return False
            # check number of affected individuals
            numAff = sum([self.pedigree.indByID(id).affected() for id in offID + offSpouseID + grandOffID]) + father.affected() + mother.affected()
            if numAff < self.numOfAffected[0] or numAff > self.numOfAffected[1]:
                return False
            return True
        # find all families with at least minOffFields...
        if not isSequence(self.families):
            self.selectedIDs = list(filter(qualify, self.pedigree.individualsWithRelatives(['spouse'] + minOffFields + minGrandOffFields)))
        else:
            self.selectedIDs = []
            for sp in range(self.pedigree.numSubPop()):
                self.selectedIDs.append(list(filter(quality, self.pedigree.individualsWithRelatives(['spouse'] + minOffFields + minGrandOffFields, subPops=sp))))


def drawThreeGenFamilySample(pop, families, numOffspring, pedSize, numOfAffected=0,
    subPops=ALL_AVAIL, idField='ind_id', fatherField='father_id', motherField='mother_id'):
    '''Draw three-generation families from a population. Such families consist
    of grant parents, their children, spouse of these children, and grand
    children. Number of offspring, total number of individuals, and total
    number of affected individuals in a pedigree should be specified using
    parameters ``numOffspring``, ``pedSize`` and ``numOfAffected``, which can all
    be a single number, or a range ``[a, b]`` (``b`` is incldued). If a single
    ``families`` is given, Pedigrees are drawn randomly from the whole
    Population or from specified (virtual) subpopulations (parameter
    ``subPops``). Otherwise, a list of numbers should be used to specify
    numbers of families from each subpopulation, which can be all
    subpopulations if ``subPops=ALL_AVAIL`` (default), or from each of the
    specified (virtual) subpopulations. This function returns a population that
    contains extracted individuals.
    '''
    return ThreeGenFamilySampler(families, numOffspring, pedSize, numOfAffected,
        subPops, idField, fatherField, motherField).drawSample(pop)
 

def drawThreeGenFamilySamples(pop, families, numOffspring, pedSize, numOfAffected=0,
    numOfSamples=1, subPops=ALL_AVAIL, idField='ind_id', fatherField='father_id',
    motherField='mother_id'):
    '''Draw ``numOfSamples`` three-generation pedigree samples from population ``pop``
    and return a list of populations. Please refer to function
    ``drawThreeGenFamilySample`` for a description of other parameters.
    '''
    return ThreeGenFamilySampler(families, numOffspring, pedSize, numOfAffected,
        subPops, idField, fatherField, motherField).drawSamples(pop, numOfSamples)


class CombinedSampler(BaseSampler):
    '''A combined sampler accepts a list of sampler objects, draw samples and
    combine the returned sample into a single population. An id field is
    required to use this sampler, which will be used to remove extra copies of
    individuals who have been drawn by different samplers.
    '''
    def __init__(self, samplers=[], idField='ind_id'):
        '''
        samplers
            A list of samplers
        '''
        BaseSampler.__init__(self)
        self.samplers = samplers
        self.idField = idField

    def drawSample(self, pop):
        allIDs = []
        self.pop = pop.clone()
        for s in self.samplers:
            sample = s.drawSample(self.pop)
            IDs = []
            for gen in range(sample.ancestralGens() + 1):
                sample.useAncestralGen(gen)
                IDs.extend(sample.indInfo(self.idField))
            allIDs.extend(IDs)
            self.pop.removeIndividuals(IDs=IDs, idField=self.idField)
        #
        # extract these guys from the original population
        return pop.extractIndividuals(IDs = allIDs, idField = self.idField)


def drawCombinedSample(pop, samplers, idField='ind_id'):
    '''Draw different types of samples using a list of ``samplers``. A
    Population consists of all individuals from these samples will
    be returned. An ``idField`` that stores an unique ID for all individuals
    is needed to remove duplicated individuals who are drawn multiple
    numOfSamples from these samplers.
    '''
    return CombinedSampler(samplers, idField=idField).drawSample(pop)

def drawCombinedSamples(pop, samplers, numOfSamples=1, idField='ind_id'):
    '''Draw combined samples ``numOfSamples`` numOfSamples and return a list of populations.
    Please refer to function ``drawCombinedSample`` for details about
    parameters ``samplers`` and ``idField``.
    '''
    return CombinedSampler(samplers, idField=idField).drawSamples(pop, numOfSamples)

